from typing import Iterable, List, MutableMapping, Tuple, Any
from itertools import groupby
import dataclasses
from functools import lru_cache

import networkdisk as nd
from pydantic import BaseModel

from vizitig import compatible_versions, version
from vizitig.errors import (
    IncompatibleViziGraph,
    NotAViziGraphError,
    MetadataDuplicate,
    MetaNotFound,
)
from vizitig.types import (
    AvailableKmerTypes,
    ESign,
    SubseqData,
    ViziKey,
    Zero,
    encode_kmer,
    decode_kmer,
    Metadata,
    metadata_from_dict,
)

from vizitig.utils import vizitig_logger


class NodeDesc(BaseModel):
    seq: str
    metadatas: List[Tuple[Metadata, Any]]
    neighbors: dict[int, ESign | None] = {}


class GraphMetadata(BaseModel):
    k: int
    size: int
    name: str
    edge_size: int  # number of edges
    gid: str  # a uniq global identifier of the graph generated once
    states: dict[str, str] = dict()
    types_list: List[str] = list()
    filter_list: MutableMapping[str, str] = dict()
    vizitig_version: str = version
    kmer_size: int | None

    def __hash__(self):
        return hash((type(GraphMetadata).__name__, self.name))

    def model_post_init(self, __context):
        nd.utils.serialize.encoderFunctions[self.key_type] = (
            self.encoder,
            self.decoder,
        )

    @property
    def graph(self) -> nd.sqlite.Graph | nd.sqlite.DiGraph:
        from vizitig.info import get_graph

        return get_graph(self.name)

    @property
    def metadata_graph(self) -> nd.sqlite.DiGraph:
        return self.graph.metadata_graph

    @property
    def key_type(self):
        return f"vizi_{self.gid}".upper()

    @lru_cache(maxsize=None)
    def encoder(self, to_encode: ViziKey | Zero) -> int | float | bytes:
        if isinstance(to_encode, float):
            return to_encode
        if isinstance(to_encode, Metadata):
            if to_encode.offset is None:
                to_encode = self.get_metadata(to_encode.type, to_encode.id)
                assert to_encode.offset is not None
            return -(to_encode.offset + 10)
        if isinstance(to_encode, AvailableKmerTypes):
            return encode_kmer(to_encode, self.k)
        if to_encode == "sequence":
            return -1
        if to_encode == "occurence":
            return -2

        if "Kmer" in str(type(to_encode)):
            raise TypeError(f"""your Kmer Class ({type(to_encode)})is locally unknown. 
                                Avoid if this error occurs during tests.""")
        if isinstance(to_encode, Zero):
            return 0
        raise NotImplementedError(f"to_encode is of incorrect type {type(to_encode)}")

    def add_filter(self, fname: str, filter: str):
        assert isinstance(filter, str)
        self.filter_list[fname] = filter

    def get_filters(self) -> Iterable[tuple[str, str]]:
        return self.filter_list.items()

    def remove_filter(self, fname: str):
        assert fname in self.filter_list
        self.filter_list.pop(fname)

    @lru_cache(maxsize=None)
    def decoder(self, to_decode):
        if isinstance(to_decode, bytes) or (
            isinstance(to_decode, int) and to_decode >= 0
        ):
            return decode_kmer(to_decode, self.k)
        if isinstance(to_decode, float):
            return to_decode
        if to_decode == -1:
            return "sequence"
        if to_decode == -2:
            return "occurence"
        if to_decode <= -3:
            return self.get_metadata_offset(-to_decode - 10)

        raise NotImplementedError(
            f"to_decode is of incorrect type or value {to_decode, type(to_decode)}",
        )

    def add_iterative_metadatas(
        self,
        iterator: Iterable[Metadata],
    ) -> Iterable[Metadata]:
        grouped_metadata = groupby(
            sorted(iterator, key=lambda e: (e.type, e.id)),
            key=lambda e: (e.type, e.id),
        )
        meta_to_add = []
        meta_to_get = []
        all_metas = {key: d for key, d in self.metadata_graph.nodes(data=True)}
        offset = len(all_metas)
        assert self.metadata_graph.graph["size"] == offset
        for (typ, id_), group in grouped_metadata:
            versions = list(group)

            short_repr = versions[0].short_repr()
            meta_to_get.append(short_repr)
            if short_repr in all_metas:
                continue
            m = dataclasses.asdict(versions[0])
            if isinstance(versions[0], SubseqData):
                start = m["start"]
                stop = m["stop"]
                for vers in versions[1:]:
                    assert isinstance(vers, SubseqData)
                    start = min(vers.start, start)
                    stop = max(vers.stop, stop)
                    for key in ("gene", "transcript"):
                        if key in m:
                            m[key] = m[key] or getattr(vers, key)
                m["start"] = start
                m["stop"] = stop

            if typ not in self.types_list:
                self.types_list.append(typ)
            m["offset"] = offset
            offset += 1
            meta_to_add.append((short_repr, m))
        self.metadata_graph.add_nodes_from(meta_to_add)
        all_metas.update({k: d for k, d in meta_to_add})
        res = list(metadata_from_dict(all_metas[k]) for k in meta_to_get)
        self.metadata_graph.graph["size"] = len(all_metas)
        self.commit_to_graph()
        return res

    def add_metadata(self, m_to_add: Metadata) -> Metadata:
        return next(iter(self.add_iterative_metadatas([m_to_add])))

    def set_all_offsets(self):
        nodes_without_offset = self.metadata_graph.find_all_nodes("offset", offset=None)
        size = self.metadata_graph.graph["size"]
        data = [
            (node, dict(offset=size + 1 + i))
            for i, node in enumerate(nodes_without_offset)
        ]
        self.metadata_graph.add_node_data_from(data)
        self.metadata_graph.graph["size"] = size + len(data)

    def get_metadata(self, type: str, id: str) -> Metadata:
        iq = self.metadata_graph.find_all_nodes("type", type=type)
        iq = iq.intersection(self.metadata_graph.find_all_nodes("id", id=id))
        nodes = list(iq)
        if len(nodes) > 1:
            raise MetadataDuplicate(f"{nodes[0]} founds with {len(nodes)} copy")
        if not nodes:
            vizitig_logger.warning(f"Not node returned for Meta {type}{id}. Skipping")
            # return SubseqData()
            raise MetaNotFound(f"{type}({id})")
        return metadata_from_dict(self.metadata_graph.nodes[nodes[0]].fold())

    def all_metadata_by_types(self, type: str) -> Iterable[Metadata]:
        select_metadata = self.metadata_graph.find_all_nodes("type", type=type)
        iq = self.metadata_graph.nbunch_iter(select_metadata, data=True)
        return list(iq.map(lambda e: metadata_from_dict(e[1])))

    def get_metadata_offset(self, offset: int) -> Metadata:
        try:
            node = self.metadata_graph.find_one_node("offset", offset=offset)
            return metadata_from_dict(self.metadata_graph.nodes[node].fold())
        except nd.exception.NetworkDiskError:
            raise MetaNotFound(offset)

    def get_all_by_type(self, type: str) -> Iterable[Metadata]:
        iq = self.metadata_graph.find_all_nodes("type", type=type)
        L = list(self.metadata_graph.subgraph(iq).nodes(data=True))
        return [metadata_from_dict(e) for _, e in L]

    def commit_to_graph(
        self,
        G: nd.sqlite.Graph | nd.sqlite.DiGraph | None = None,
        set_offset: bool = True,
    ) -> None:
        if G is None:
            G = self.graph
        G.graph = self.model_dump()

    def get_states(self) -> list[str]:
        return list(self.states)

    def get_state(self, state_name: str) -> str:
        return self.states[state_name]

    def set_state(self, state_name: str, state: str):
        self.states[state_name] = state
        self.commit_to_graph()

    def delete_state(self, state_name: str):
        del self.states[state_name]
        self.commit_to_graph()

    @classmethod
    def set_metadata(
        cls,
        G: nd.sqlite.DiGraph | nd.sqlite.Graph,
        check_compatibility: bool = True,
        name: str | None = None,
    ):
        d = G.graph.fold()
        d.setdefault("name", name)
        if "vizitig_version" not in d:
            raise NotAViziGraphError()
        if check_compatibility and d["vizitig_version"] not in compatible_versions:
            raise IncompatibleViziGraph(d["vizitig_version"])
        GM = cls(**d)
        assert GM.gid
        G.metadata = GM

    def to_nodedesc(
        self,
        d: dict[ViziKey, None | Any],
        neighbors: dict[int, ESign | None],
    ) -> NodeDesc:
        metadatas = list()
        for k, v in d.items():
            if isinstance(k, Metadata):
                metadatas.append((k, v))
        return NodeDesc(
            seq=str(d["sequence"]),
            metadatas=metadatas,
            neighbors=neighbors,
        )



import asyncio
import concurrent.futures


def run_with_keyword(f, *args):
    return f(*args[:-1], **args[-1])


def async_subproc(fct):
    """In async context, we need to decorate long running
    process (and CPU intensive one) to be able to fork them
    appropriately.

    For some weird reason, run_in_executor takes only *args and not
    kword argument.

    """

    async def _fct(*args, **kwargs):
        loop = asyncio.get_running_loop()
        with concurrent.futures.ProcessPoolExecutor() as pool:
            return await loop.run_in_executor(
                pool, run_with_keyword, fct, *args, kwargs
            )

    return _fct



from pathlib import Path
from vizitig.env_var import VIZITIG_DIR

root = Path(VIZITIG_DIR)

if not root.exists():
    root.mkdir()

graphs_path = root / "data"
graphs_log = root / "log"
index_path = root / "index"

if not graphs_path.exists():
    graphs_path.mkdir()


def graph_path_name(name: str) -> Path:
    return Path(graphs_path, f"{name}.db")


def log_path_name(name: str) -> Path:
    if not graphs_log.exists():
        graphs_log.mkdir()
    return Path(graphs_log, f"{name}.db")


def index_path_name(name: str, create=True) -> Path:
    path = index_path / name
    if not path.exists() and create:
        path.mkdir(parents=True)
    return path



from pathlib import Path
from typing import Iterator, Tuple

from vizitig.cli import subparsers
from vizitig.info import get_graph, reload_graph, GraphLogger
from vizitig.parsing import parse_annotations_for_genes, parse_reference_sequences_dna
from vizitig.types import DNA, SubseqData
from vizitig.utils import vizitig_logger as logger
from vizitig.color.bulk_update import bulk_annotate_graph


def main(args):
    with GraphLogger(args.graph_name, "gene"):
        gene(args.graph_name, args.metadata, args.ref_seq, args.parse_only)


def gene(
    graph_name: str, metadata_file: Path, ref_seq_file: Path, parse_only: bool = False
) -> None:
    Graph = get_graph(graph_name)
    k = Graph.metadata.k
    annotation_data: dict[str, SubseqData] = {
        s.id: s for s in parse_annotations_for_genes(metadata_file, k)
    }
    for subseq, _ in parse_reference_sequences_dna(
        ref_seq_file, Graph.metadata, annotation_data
    ):
        Graph.metadata.add_metadata(subseq)

    def custom_generator() -> Iterator[Tuple[DNA, int]]:
        for subseq, dna in parse_reference_sequences_dna(
            ref_seq_file, Graph.metadata, annotation_data
        ):
            yield (dna, Graph.metadata.encoder(subseq))

    if parse_only:
        it_dna = custom_generator()
        it_kmer = (
            (kmer, meta) for dna, meta in it_dna for kmer in dna.enum_canonical_kmer(k)
        )
        logger.info("parsing only")
        logger.info(f"count {sum(1 for _ in it_kmer)}")
        return
    Graph.metadata.commit_to_graph()
    bulk_annotate_graph(graph_name, custom_generator)
    Graph.metadata.commit_to_graph()
    reload_graph(graph_name)


parser = subparsers.add_parser(
    "genes",
    help="Add gene annotation to a given graph",
)

parser.add_argument(
    "graph_name",
    help="A graph name. List possible graph with python3 -m vizitig info",
    metavar="graph",
    type=str,
)

parser.add_argument(
    "-r",
    "--ref-seq",
    help="Path toward a (possibly compressed) fasta files containing reference sequences",
    metavar="refseq",
    type=Path,
    required=True,
)

parser.add_argument(
    "-m",
    "--metadata",
    help="Path towards a (possibly compressed) gtf files containing metadata of reference sequences",
    metavar="gtf",
    type=Path,
    required=True,
)

parser.add_argument(
    "-p",
    "--parse-only",
    help="Go through the data without ingesting. Provides the number of line to ingest",
    action="store_true",
)

parser.set_defaults(func=main)



from pathlib import Path

from vizitig.info import get_graph
from vizitig.index import load_kmer_index
from vizitig.errors import FixtureError


def test_fixture_setup_correctly(tmp_path: Path, viz_dir, test_graph_name):
    try:
        assert get_graph(test_graph_name)
        assert load_kmer_index(test_graph_name)
    except Exception as e:
        raise FixtureError(
            "The following issue occured with the fixture used for testing : {}. Please contact the developpers.".format(
                e
            )
        )



from pytest import fixture
from subprocess import run, PIPE
from pathlib import Path
import os
import shutil

from vizitig.env_var import VIZITIG_SHORT_TEST
from vizitig.utils import vizitig_logger as log
from vizitig.info import delete_graph


@fixture(scope="session")
def viz_dir():
    vizitig_dir = Path(os.environ["VIZITIG_DIR"])
    if vizitig_dir.exists():
        shutil.rmtree(vizitig_dir)
    run("make small_ex", stdout=PIPE, stderr=PIPE, shell=True)
    if not VIZITIG_SHORT_TEST:
        run("make small_ex_alt", stdout=PIPE, stderr=PIPE, shell=True)
        run("make small_ex_alt2", stdout=PIPE, stderr=PIPE, shell=True)
    else:
        log.warning("SHORT TEST ONLY")
    yield vizitig_dir
    shutil.rmtree(vizitig_dir)


def reset_graph():
    delete_graph("mini_bcalm")
    if not VIZITIG_SHORT_TEST:
        delete_graph("mini_bcalm_alt1")
        delete_graph("mini_bcalm_alt2")

    run("make small_ex", stdout=PIPE, stderr=PIPE, shell=True)

    if not VIZITIG_SHORT_TEST:
        run("make small_ex_alt", stdout=PIPE, stderr=PIPE, shell=True)
        run("make small_ex_alt2", stdout=PIPE, stderr=PIPE, shell=True)
    else:
        log.warning("SHORT TEST ONLY")


@fixture(scope="function")
def test_graph_name():
    return "mini_bcalm"


@fixture(scope="function")
def kmer_set_and_nodes():
    return [
        ("ATCGTGAGTCGTAGCTGATGC", [1]),
        ("TTATTCGATTAGCAGTTAGCT", [2]),
        ("AGTCGGATCGATAGCTGATAG", [3]),
    ]


@fixture(scope="function")
def sequences_set():
    return [
        ("ATCGTGAGTCGTAGCTGATGCTAGCTGATCGATCGGATGTCGTAGCATCGNATTCCAaaaC", [1]),
        ("AGTTATTCGATTAGCAGTTAGCT".lower, [2]),
        ("AGTCGGATCGATAGCTGATAGCTAGCT", [3]),
    ]



import gzip
from pathlib import Path
from typing import Callable, Iterable, Iterator, cast

from vizitig.cli import subparsers
from vizitig.errors import KmerIndexNeeded, NoIndex
from vizitig.export import export_format, export_graph
from vizitig.index import load_kmer_index
from vizitig.info import are_kmer_ingested, get_graph, graph_info, graphs_list
from vizitig.parsing import parse_fasta_dna
from vizitig.types import DNA, Kmer
from vizitig.utils import SubLog, progress
from vizitig.utils import vizitig_logger as logger
from vizitig.index import temporary_kmerset
from vizitig.env_var import VIZITIG_NO_TMP_INDEX

export_formats_str = ", ".join(sorted(export_format))


def fetch_nodes_by_dna(
    graph_name: str, dnas: Callable[[], Iterator[DNA]]
) -> Iterable[int]:
    """Will fetch the nodes that contain one of the provided kmers:"""
    G = get_graph(graph_name)
    k = G.metadata.k
    try:
        kmer_index = load_kmer_index(graph_name)
        logger.info(f"found index {type(kmer_index).__name__}")
        if not VIZITIG_NO_TMP_INDEX:
            index_set = temporary_kmerset(
                dnas,
                k=G.metadata.k,
                index_type=kmer_index.index_type.__name__,
                shard_number=kmer_index.shard_number,
            )
            logger.info("index built")
            return kmer_index.intersection_index(index_set)
    except NoIndex:
        logger.warning("no index found, fall back to ingested kmer")

    def kmers():
        return (kmer for dna in dnas() for kmer in dna.enum_canonical_kmer(k))

    return fetch_nodes_by_kmers(graph_name, kmers)


def fetch_nodes_by_kmers(
    graph_name: str, kmers: Iterator[Kmer] | Callable[[], Iterator[Kmer]]
) -> Iterable[int]:
    """Will fetch the nodes that contain one of the provided kmers"""
    G = get_graph(graph_name)
    try:
        kmer_index = load_kmer_index(graph_name)
        logger.info(f"found index {type(kmer_index).__name__}")
        if not VIZITIG_NO_TMP_INDEX:
            kmers = cast(Callable[[], Iterator[Kmer]], kmers)
            index_set = temporary_kmerset(
                kmers,
                k=G.metadata.k,
                index_type=kmer_index.index_type.__name__,
                shard_number=kmer_index.shard_number,
            )
            return kmer_index.intersection_index(index_set)
        else:
            if not hasattr(kmers, "__iter__"):
                kmers = cast(Callable[[], Iterator[Kmer]], kmers)
                return kmer_index.intersection(kmers())

            kmers = cast(Iterator[Kmer], kmers)
            return kmer_index.intersection(kmers)

    except NoIndex:
        logger.warning("no index found, fall back to ingested kmer")
    if not are_kmer_ingested(graph_name):
        raise KmerIndexNeeded()
    ## First we store kmers in a temporary table
    KmerTempTable = G.schema.schema.add_table("tmp_kmer", temporary=True)
    kmer_col = KmerTempTable.add_column("kmer", sqltype=G.schema.nodes[1].sqltype)
    G.helper.execute(KmerTempTable.create_query())
    q = KmerTempTable.insert_many()
    with SubLog("Indexing"):
        G.helper.execute(KmerTempTable.create_index_query((kmer_col,)))
    with SubLog("ingestion"):
        G.helper.executemany(q, map(lambda e: (e,), progress(kmers)))
    return G.find_all_nodes(lambda e: e.inset(KmerTempTable))


def fetch_nodes_by_gene(graph_name: str, gene: str) -> set[int]:
    """Will fetch all the nodes labeled by one gene"""
    G = get_graph(graph_name)
    return G.find_all_nodes(gene)


def main(args) -> None:
    info = graph_info(args.graph_name)
    k = info["k"]
    nodes_ids: set[int] = set()
    for kmer_file in args.kmer_files:
        openfile: Callable = open
        if kmer_file.suffix == ".gz":
            openfile = gzip.open

        def generator() -> Iterator[DNA]:
            with openfile(kmer_file) as f:
                yield from parse_fasta_dna(f, k)

        nodes_ids.update(fetch_nodes_by_dna(args.graph_name, generator))
    for node_file in args.node_files:
        with open(node_file) as f:
            nodes_ids.update(map(int, f.read().strip().split()))
    dnas = list(map(lambda e: DNA(e), args.dnas))
    nodes_ids.update(
        fetch_nodes_by_dna(
            args.graph_name,
            lambda: iter(iter(dnas)),
        )
    )
    nodes_ids.update(args.nodes_ids)
    for gene in args.genes:
        nodes_ids.update(fetch_nodes_by_gene(args.graph_name, gene))
    logger.info(f"Collected nodes: {len(nodes_ids)}")
    export_graph(args.graph_name, nodes_ids, args.export_format, args.output_file)


parser = subparsers.add_parser(
    "search",
    description=f"""
Find nodes according to various criteria and output the subgraph in various format.

The nodes can be selected by several means: (kmers in a fasta, unitig ids, genes).
If several options are selected, the results will be cumulative (a node is
selected if it satisified any of the condition.)

Result can be exported to any of the following format: {export_formats_str}.
The default is simply a node list. Most of the export relies on networkx to do the work.

- Pickle will allow to load the networkx graph in Python directly.
- Json is the JSON serialisation of networkx
- bcalm is a custom implementation of the bcalm format with extra metadata provided in the headers (genes and colors)

- For GraphViz format, if a non-dot extension is used, drawing to a file will be attempted.
Requires pygraphviz installed (non-default as it requires also graphviz binaries and dev
to be installed).
""",
)

parser.add_argument(
    "graph_name",
    help=f"A graph name. {graphs_list()}",
    metavar="graph",
    type=str,
    choices=graphs_list(),
)

parser.add_argument(
    "-o",
    "--output-file",
    help="File to store the result. Default is stdin.",
    type=Path,
    required=True,
)

parser.add_argument(
    "-f",
    "--export-format",
    help=f"The export format. One of {export_formats_str}. Default is nodelist",
    default="nodelist",
)

parser.add_argument(
    "-kf",
    "--kmer-files",
    help="Path toward files containing kmers (fasta format, possibly gzip compressed)",
    metavar="kmer_files",
    nargs="+",
    default=[],
    type=Path,
)

parser.add_argument(
    "-d",
    "--dna",
    help="Enumeration of DNA sequences",
    metavar="dna",
    nargs="+",
    default=[],
    type=str,
)

parser.add_argument(
    "-n",
    "--nodes-ids",
    help="Explicit list of node id",
    metavar="nodes",
    nargs="+",
    default=[],
    type=int,
)

parser.add_argument(
    "-nf",
    "--node-files",
    help="A path toward a file containing a list of node id (one per line)",
    metavar="nodes",
    default=[],
    nargs="+",
    type=Path,
)

parser.add_argument(
    "-g",
    "--genes",
    nargs="+",
    help="A list of gene annotations",
    type=str,
    default=[],
)


parser.set_defaults(func=main)



from pytest import fixture, mark

from vizitig.info import get_graph
from vizitig.index import (
    load_kmer_index,
    index_info,
    build_kmer_index,
    temporary_kmerindex,
    temporary_kmerset,
)
from vizitig.index import index_types, Shard
from vizitig.types import DNA

from random import choices


from tempfile import TemporaryDirectory
from pathlib import Path


@fixture(scope="session")
def index_dir():
    tmp = TemporaryDirectory()
    path = Path(tmp.name)
    yield path
    tmp.cleanup()


@fixture(scope="session")
def dna():
    return DNA("".join(choices("ACGT", k=100)))


@fixture(scope="session")
def dnas():
    """Generate DNA sequence with number without repeating min_k-mer"""
    res = []
    kmer = set()
    for i in range(20):
        dna = DNA("".join(choices("ACGT", k=i**2)))
        nkmer = set(dna.enum_canonical_kmer(min_k))
        if not kmer.intersection(nkmer):
            res.append((dna, i))
        kmer.update(nkmer)
    return tuple(res)


valid_k = [2, 7, 10, 11, 21, 31, 33, 55, 63, 127, 511]


@mark.parametrize("IndexType", [Shard.subclasses[e] for e in index_types])
@mark.parametrize("k", valid_k)
def test_build_dna(index_dir, IndexType, dnas, k):
    filt_dnas = []  # to filter dna seq with shared kmer
    kmers = set()
    for dna, _ in dnas:
        nkmers = set(dna.enum_canonical_kmer(k))
        if kmers.intersection(nkmers):
            continue
        kmers.update(nkmers)
        filt_dnas.append(dna)
    d_dna = {dna: i for i, dna in enumerate(filt_dnas)}
    d_kmer = {
        kmer: i
        for i, dna in enumerate(filt_dnas)
        for kmer in dna.enum_canonical_kmer(k)
    }

    path_dna = index_dir / (IndexType.__name__ + "dna")
    index_dna = IndexType.build_dna(path_dna, 0, 1, iter(d_dna.items()), k)

    path_kmer = index_dir / (IndexType.__name__ + "kmer")
    index_kmer = IndexType.build_kmer(path_kmer, 0, 1, iter(d_kmer.items()), k)

    assert dict(index_dna) == d_kmer
    assert dict(index_kmer) == d_kmer


@mark.parametrize("IndexType", [Shard.subclasses[e] for e in index_types])
@mark.parametrize("k", valid_k)
def test_all_index(index_dir, IndexType, dna, k):
    path = index_dir / IndexType.__name__
    d = dict((kmer, i) for i, kmer in enumerate(dna.enum_canonical_kmer(k)))
    index = IndexType.build_kmer(path, 0, 1, iter(d.items()), k)
    assert set(index) == set(dna.enum_canonical_kmer(k))
    assert dict(index) == d

    subset = {k: i for k, i in d.items() if i % 2 == 0}
    # join_iter
    expected_res = {i: i for i in subset.values()}
    obtained_res = index.join(iter(subset.items()))
    assert dict(obtained_res) == expected_res

    # intersection_iter
    expected_res_intersection = {i for i in subset.values()}
    obtained_res_intersection = index.intersection(iter(subset))
    assert set(obtained_res_intersection) == expected_res_intersection

    subset_2 = {k: i**2 for k, i in d.items() if i % 2 == 0}
    # join_index:
    path2 = Path(path.parent, path.name + "2")
    index2 = IndexType.build_kmer(path2, 0, 1, iter(subset_2.items()), k)
    assert dict(index2) == subset_2
    expected_res_join_index = {i: i**2 for i in d.values() if i % 2 == 0}
    obtained_res_join_index = index.join_index(index2)
    assert dict(obtained_res_join_index) == expected_res_join_index

    # intersection_index
    path3 = Path(path.parent, path.name + "3")
    index3 = IndexType.Set.build_kmer(path3, 0, 1, iter(subset_2), k)
    assert set(index3) == set(subset_2)

    expected_res_join_index = {i for i in d.values() if i % 2 == 0}
    obtained_res_join_index = index.intersection_index(index3)
    assert set(obtained_res_join_index) == expected_res_join_index


@mark.parametrize("IndexType", [Shard.subclasses[e] for e in index_types])
@mark.parametrize("k", valid_k)
def test_join_val_duplicate(index_dir, IndexType, dna, k):
    path = index_dir / (IndexType.__name__ + "_join")
    d = {kmer: i % 3 for i, kmer in enumerate(dna.enum_canonical_kmer(k))}
    index = IndexType.build_kmer(path, 0, 1, iter(d.items()), k)
    assert set(index) == set(dna.enum_canonical_kmer(k))
    assert dict(index) == d

    path2 = index_dir / (IndexType.__name__ + "_join2")
    d2 = {kmer: i for i, kmer in enumerate(dna.enum_canonical_kmer(k))}
    index2 = IndexType.build_kmer(path2, 0, 1, iter(d2.items()), k)
    assert set(index) == set(dna.enum_canonical_kmer(k))
    assert dict(index2) == d2

    # join_index:
    expected_res_join = sorted(
        set((d[kmer], d2[kmer]) for kmer in dna.enum_canonical_kmer(k))
    )

    obtained_res_join_iter = sorted(index.join(d2.items()))
    assert obtained_res_join_iter == expected_res_join

    obtained_res_join_index = sorted(index.join_index(index2))
    assert obtained_res_join_index == expected_res_join


min_k = 12


@mark.parametrize("index_type", index_types)
@mark.parametrize("k", [k for k in valid_k if k > min_k])
@mark.parametrize("shard_number", [1, 5])
def test_tmp_kmerindex(index_type, dnas, k, shard_number):
    def generator():
        return iter(dnas)

    index = temporary_kmerindex(
        generator, k=k, shard_number=shard_number, index_type=index_type
    )
    expected = {
        kmer: i for dna, i in generator() for kmer in dna.enum_canonical_kmer(k)
    }
    for kmer, i in expected.items():
        assert index[kmer] == i

    assert len(index) == len(expected)


@mark.parametrize("index_type", index_types)
@mark.parametrize("k", [k for k in valid_k if k > min_k])
@mark.parametrize("shard_number", [1, 5])
def test_tmp_kmerset(index_type, dnas, k, shard_number):
    def generator():
        return (d for d, _ in dnas)

    index = temporary_kmerset(
        generator, k=k, shard_number=shard_number, index_type=index_type
    )
    expected = {kmer for dna in generator() for kmer in dna.enum_canonical_kmer(k)}
    for kmer in expected:
        assert kmer in index
    assert len(index) == len(expected)


def test_fixture_setup_test_graph(viz_dir, test_graph_name):
    G = get_graph(test_graph_name)
    indexes = index_info(G.name)
    assert len(indexes) > 0


def test_drop_index(viz_dir, test_graph_name):
    G = get_graph(test_graph_name)
    inst_index_info = index_info(G.name)
    index_type = inst_index_info[0].type
    index_info_size = len(inst_index_info)
    index = load_kmer_index(G.name, index_type)
    assert index is not None
    index.drop()
    assert len(index_info(G.name)) == (index_info_size - 1)
    build_kmer_index(G.name, index_type)  # to restore the situation



import argparse

parser = argparse.ArgumentParser(
    prog="vizitig",
    description="A CLI interface to vizitig",
    formatter_class=argparse.ArgumentDefaultsHelpFormatter,
)
parser.set_defaults(func=lambda e: e)

subparsers = parser.add_subparsers(required=True)



from fastapi import FastAPI
from fastapi.staticfiles import StaticFiles

from vizitig.api.main import app, export_path
from vizitig.api.path import site
from vizitig.info import graphs_path

"""
Vizitig API
===========

This module is based on FastApi to do most of the work. Code in this module should
be mostly:
- boilerplate code to call the appropriate vizitig function
- a bit of tmpfile manipulation

API Layout
----------

The api code is mostly in the main submodule.
The api must serves some files (like the website directory)
that are provided here. 
To add a completely novel api endpoint, do it here exactly like for the export
and site endpoint. 

Specificities
-------------

FastApi decorator are overloaded to add the Python function name in the openapi signature.
Thanks to that, the JS front API client use the same function name.
As a consequence, instead of defining a route using

app.get("/foo/bar")

use the customly defined in the main submodule

vizitig.main.get("/foo/bar")

Be aware that the JS code might not handle so well optional argument or complexe interaction.


Long running process
--------------------

For long running vizitig function, wrap the function using async_subproc.
Otherwise, the API endpoint might be too busy to serve queries while doing
heavy tasks. Ex:

```python
from vizitig.api.async_utils import async_subproc
async_subproc(generate_graph)(tmp_path, graph_path_name(name))
````

Error Handling
--------------

Errors raise by Vizitig (VizitigException) are catched and lift to the front.
Errors raise by anything else will not be catched and raise a 500 server Error.
To avoid this, try/except those error and raise instead the appropriate HTTPException

In principle, Error 500 should never happens. If it happens, it means that some part of
the API is not handling an error appropriately. Please report the issue on the project
gitlab if it is the case.

"""


main_api = FastAPI()
main_api.mount("/api", app)
main_api.mount("/files", StaticFiles(directory=graphs_path, html=True), name="files")
main_api.mount("/export", StaticFiles(directory=export_path, html=True), name="export")
main_api.mount("/", StaticFiles(directory=site, html=True), name="site")



from pathlib import Path

from pytest import raises, mark

from vizitig.errors import EmptyExportError
from vizitig.export import (
    export_graph,
    export_bcalm_by_iterator,
    dump_one_node_bcalm,
    dump_one_node_bcalm_for_iterator,
)
from vizitig.info import get_graph

formats = [
    "json",
    "bcalm",
    "nodelist",
]  # "networkx_pickle" and "gml" should be in this list


@mark.parametrize("format", formats)
def test_export_graph_bcalm_in_normal_conditions(
    tmp_path: Path, viz_dir, test_graph_name, format: str
):
    test_file = tmp_path / "file"
    export_graph(test_graph_name, [165114, 248683, 459264], format, test_file)
    assert test_file.exists()
    test_file.unlink()


@mark.parametrize("format", formats)
def test_export_graph_bcalm_subgraph(
    tmp_path: Path, viz_dir, test_graph_name, format: str
):
    test_file = tmp_path / "file"
    sample_nodes = [165114, 248683, 459264]

    export_graph(test_graph_name, sample_nodes, format, test_file)
    assert test_file.exists()
    test_file.unlink()


@mark.parametrize("format", formats)
def test_export_graph_bcalm_crashes_with_bad_nodes_id(
    tmp_path: Path, viz_dir, test_graph_name, format: str
):
    test_file = tmp_path / "file"
    with raises(EmptyExportError):
        export_graph(test_graph_name, list([-1]), format, test_file)
    assert not test_file.exists()


def test_export_graph_by_iterator(tmp_path: Path, viz_dir, test_graph_name):
    test_file = tmp_path / "file"
    export_bcalm_by_iterator(test_graph_name, test_file)
    assert tmp_path.exists()
    test_file.unlink()


def test_dump_one_node_bcalm_for_iterator(tmp_path: Path, viz_dir, test_graph_name):
    G = get_graph(test_graph_name)
    for node in G._node:
        dumped = dump_one_node_bcalm(G, node)
        assert isinstance(dumped, str)
        assert dumped[0] == ">"
        assert "\n" in dumped
        assert dumped[::-1][0] in ["A", "T", "C", "G"]



from pathlib import Path
from typing import Tuple, Iterator, Literal, TypeAlias, get_args, List

from vizitig.cli import subparsers
from vizitig.info import get_graph, reload_graph
from vizitig.parsing import (
    parse_annotations,
    parse_reference_sequence_for_full_annotation,
    parse_transcript_or_exon_refseq,
)
from vizitig.errors import ParseError, FormatInputError
from vizitig.types import DNA, SubseqData
from vizitig.color.bulk_update import bulk_annotate_graph
from vizitig.utils import SubLog
from vizitig.utils import vizitig_logger as logger, progress
import re


def annotate_genome(
    graph_name: str, metadata: Path, ref_seq: Path, select_only: List[str] | None = None
):
    with SubLog("Annotate annotate"):
        Graph = get_graph(graph_name)
        k = Graph.metadata.k

        if select_only is not None:
            annotation_data: list[SubseqData] = list(  # type: ignore
                elem
                for elem in parse_annotations(metadata, k)
                if elem.type.lower() in select_only
            )
        else:
            annotation_data: list[SubseqData] = list(  # type: ignore
                elem for elem in parse_annotations(metadata, k)
            )

        l: int = len(annotation_data)
        logger.info("Found {} annotations. Ingesting annotations.".format(l))

        annotation_data = Graph.metadata.add_iterative_metadatas(
            progress(annotation_data, l)
        )

        logger.info("Metadata were ingested. Procceding to annotate.")

        if not annotation_data:
            raise ParseError("No metadata found in the GFT or GFF file")

        def generator() -> Iterator[Tuple[DNA, int]]:
            it = parse_reference_sequence_for_full_annotation(
                ref_seq,
                annotation_data,
                k,
            )
            
            for meta, dna in it:
                if len(dna) > k:
                    yield (dna, Graph.metadata.encoder(meta))

        bulk_annotate_graph(graph_name, generator)

        Graph.metadata.commit_to_graph()
        reload_graph(graph_name)


annotate_metadata_type: TypeAlias = Literal["Exon", "Transcript"]


def annotate_other(
    graph_name: str,
    metadata_type: str,
    refseqs_file: Path,
    annotation_file: Path | None,
    select_only: List[str] | None = None,
):
    with SubLog("Annotate annotate"):
        Graph = get_graph(graph_name)

        logger.info("Adding all metadata to graph metadata before ingestion")

        parser = parse_transcript_or_exon_refseq(refseqs_file)

        if metadata_type not in get_args(annotate_metadata_type):
            raise FormatInputError(
                f"metadata_type {metadata_type} not in {get_args(annotate_metadata_type)}"
            )

        metas = []
        for i, elem in enumerate(parser):
            if metadata_type == "Exon":
                meta: SubseqData = str_to_exon(elem, i, meta_only=True)  # type: ignore
            elif metadata_type == "Transcript":
                meta, _ = str_to_transcript(elem)
            metas.append(meta)

        if annotation_file:
            logger.info(
                "Annotating graph {} with {} reference sequences and annotation file {}.".format(
                    graph_name, metadata_type, annotation_file
                )
            )
            for meta in parse_annotations(annotation_file, Graph.metadata.k):
                if select_only is None or meta.type.lower() in select_only:
                    metas.append(meta)

        else:
            logger.info(
                "Annotating graph {} with {} reference sequences.".format(
                    graph_name, metadata_type
                )
            )

        metas = Graph.metadata.add_iterative_metadatas(metas)

        metas_dict = {m.short_repr(): -m.offset for m in metas if m.offset is not None}

        def generator() -> Iterator[Tuple[DNA, int]]:
            parser = parse_transcript_or_exon_refseq(refseqs_file)
            for i, elem in enumerate(parser):
                if metadata_type == "Exon":
                    res = str_to_exon(elem, i)
                    assert isinstance(res, tuple)
                    metadata, sequences = res
                elif metadata_type == "Transcript":
                    metadata, sequences = str_to_transcript(elem)
                else:
                    raise NotImplementedError
                # enc = Graph.metadata.encoder(metadata)
                for s in sequences:
                    yield (s, -metas_dict[metadata.short_repr()])

        bulk_annotate_graph(graph_name, generator, metas=metas)
        reload_graph(graph_name)


def main(args) -> None:
    assert args.graph_name
    select_only: List[str] | None = (
        [elem.lower() for elem in args.select_only] if args.select_only else None
    )

    if args.genome:
        try:
            assert Path(args.metadata)
        except AssertionError:
            logger.warning(
                """
                You are trying to annotate a graph with a refseq but without metadata.
                Consider using vizitig color instead. 
                Use vizitig -h, vizitig color -h or vizitig annotate -h to get help. 
            """
            )
        logger.info(
            "Annotating graph {} with reference sequence and annotation file".format(
                args.graph_name
            )
        )

        annotate_genome(
            args.graph_name, Path(args.metadata), Path(args.genome), select_only
        )
        return

    if (args.exons and args.transcripts) or not (args.exons or args.transcripts):
        raise FormatInputError(
            "Incorrect arguments. Please choose between Transcripts or Exons ingestion."
        )

    if args.exons:
        refseqs_file = args.exons
        metadata_type = "Exon"
    elif args.transcripts:
        refseqs_file = args.transcripts
        metadata_type = "Transcript"

    if args.metadata:
        annotation_file = args.metadata
        logger.info(
            "Found annotation file. Annotations will be added with the reference sequences."
        )
    else:
        annotation_file = None
    annotate_other(
        args.graph_name, metadata_type, refseqs_file, annotation_file, select_only
    )


parser = subparsers.add_parser(
    "annotate",
    help="""
    Add annotations to a given graph. The following usages are possible :
        - If you have a genomic full reference sequence and an annotation file (gtf or gff), use 'vizitig annotate --genome path_to_genome_file --metadata path_to_annotation_file'. 
        - If you have exons or transcripts reference sequences, use 'vizitig annotate --exons path_to_exons_file'. Replace 'exons' by 'transcripts' if you have transcripts sequences. 
        
        With the later feature, you can also add --metadata path_to_annotation_file if you have an annotation file that corresponds to the exons or transcripts reference sequence, but it is not mandatory.
    
    What it does : 
        - For the genomic sequence and the annotation file, Vizitig will sort and ingest all the metadata in the annotation file and read the reference sequence using a reading frame. 
        Everytime the whole sequence of a metadata has been read, it will tag the graph with the corresponding metadata. 
        - For the exons or transcripts reference sequence, it will tag the graph with the a generic metadata that has the header of the reference sequence and later add the additional data of the annotation file. The later is therefore optional
        """,
)

parser.add_argument(
    "graph_name",
    help="A graph name. List possible graph with python3 -m vizitig info",
    metavar="graph",
    type=str,
)

parser.add_argument(
    "-g",
    "--genome",
    help="Path toward a (possibly compressed) fasta file containing a reference genome",
    metavar="genome_ref",
    type=Path,
)

parser.add_argument(
    "-m",
    "--metadata",
    help="Path towards a (possibly compressed) gtf file containing metadata of a reference sequence",
    metavar="gtf",
    type=Path,
    required=False,
)

parser.add_argument(
    "-e",
    "--exons",
    help="Path towards a fasta file containing exon reference sequences",
    type=Path,
    required=False,
)

parser.add_argument(
    "-t",
    "--transcripts",
    help="Path towards a fasta file containing transcript reference sequences",
    type=Path,
    required=False,
)

parser.add_argument(
    "-so",
    "--select_only",
    action="append",
    default=None,
    help="Select values (can be used multiple times)",
)

parser.set_defaults(func=main)


def str_to_exon(
    chunk: str, exon_offset: int, meta_only=False
) -> Tuple[SubseqData, DNA] | SubseqData:
    """
    Takes a line of exon parsing as input and returns the SubseqData and DNA corresponding to it
    """

    try:
        id = re.findall(r"\|([A-Z]{2}_.*)\|", chunk)[0]
        start = int(re.findall(":([0-9]*)-", chunk)[0])
        stop = int(re.findall("-([0-9]*)", chunk)[0])
        seq = re.findall("\n[ACTGN\n]*", chunk.upper())[0].replace("\n", "").upper()

        gene_id = re.findall(r"\(([A-Za-z0-9]*)\)", chunk)[0]

    except IndexError:
        raise ValueError("Exon parsing does not match the expected format.")

    s = SubseqData(
        id=id + "_exon_number_{}".format(exon_offset + 1),
        type="Exon",
        start=start,
        stop=stop,
        list_attr=[],
        gene=gene_id,
    )
    if meta_only:
        return s
    d = DNA.from_str(seq)
    return (s, d)


def str_to_transcript(chunk: str) -> Tuple[SubseqData, list[DNA]]:
    """
    Takes a line of transcript parsing as input and returns the SubseqData and DNA corresponding to it
    """
    match = re.match(r">(?P<id>[^\s\n]+)\s+(?P<seq>[ACGTNacgtn\n]+)", chunk)

    if match:
        s = SubseqData(
            id=match.group("id"), type="Transcript", start=0, stop=0, list_attr=[]
        )
        d = list(DNA.from_str(match.group("seq").replace("\n", "").upper()))
        return (s, d)
    else:
        raise ValueError("Transcript data does not match the expected format.")



from functools import singledispatch
from typing import Iterable

import networkdisk as nd

from vizitig.errors import InvalidKmerSize, MetaNotFound, NoIndex, QueryError
from vizitig.index import load_kmer_index, index_info
from vizitig.info import get_graph
from vizitig.query import query
from vizitig.query.query import Term, parse
from vizitig.types import DNA, Kmer
from vizitig.utils import vizitig_logger as logger


def search_kmers(
    graph_name, kmers, small_k: int | None = None
) -> nd.sql.helper.IterableQuery:
    G = get_graph(graph_name)
    kmers = set(kmers)
    if not kmers:
        return empty_iterable_query(G)
    try:
        kmer_index = load_kmer_index(graph_name, small_k=small_k)
    except NoIndex:
        small_ks = sorted((idx.k for idx in index_info(graph_name)))
        small_ks_repr = "None" if not small_ks else str(small_ks)
        raise QueryError(
            f"Index needed with small-k={small_k}, (available {small_ks_repr})"
        )
    nids = set(kmer_index.intersection(kmers))
    if nids:
        return constants_to_iterable_query(G, nids)
    return empty_iterable_query(G)


def constant_to_iterable_query(
    G: nd.sqlite.DiGraph,
    nid: int,
) -> nd.sql.helper.IterableQuery:
    col = G.dialect.columns.ValueColumn(nid)
    query = G.dialect.queries.SelectQuery(columns=(col,))
    return G.dialect.helper.IterableQuery(G.helper, query).map(lambda e: e[0])


def constants_to_iterable_query(
    G: nd.sqlite.DiGraph,
    nodes: Iterable[int],
) -> nd.sql.helper.IterableQuery:
    query = G.dialect.queries.ValuesQuery(*nodes)
    return G.dialect.helper.IterableQuery(G.helper, query).map(lambda e: e[0])


def empty_iterable_query(G: nd.sqlite.DiGraph) -> nd.sql.helper.IterableQuery:
    col = G.dialect.columns.ValueColumn(1)
    cond = G.dialect.conditions.FalseCondition()
    query = G.dialect.queries.SelectQuery(columns=(col,), condition=cond)
    return G.dialect.helper.IterableQuery(G.helper, query).map(lambda e: e[0])


def limit_iterable_query(
    iq: nd.sql.helper.IterableQuery, limit: int
) -> nd.sql.helper.IterableQuery:
    dialect = iq.helper.dialect

    return nd.sql.helper.IterableQuery(
        iq.helper, dialect.queries.SelectQuery(iq.query, limit=limit)
    ).map(lambda e: e[0])


@singledispatch
def _search(T: Term, gname: str) -> nd.sql.helper.IterableQuery:
    raise NotImplementedError(T)


@_search.register(query.Meta)
def _(T, gname):
    G = get_graph(gname)
    if T.attrs:
        raise NotImplementedError
    try:
        if T.name is None:
            metadata = list(G.metadata.get_all_by_type(T.type))
            if metadata:
                iq = G.find_all_nodes(metadata[0])
            else:
                raise MetaNotFound(T.type)
            for meta in metadata[1:]:
                iq = iq.union(G.find_all_nodes(meta))
            return iq
        meta = G.metadata.get_metadata(T.type, T.name)
    except KeyError:
        raise MetaNotFound(T)
    return G.find_all_nodes(meta)


@_search.register(query.Color)
def _(T, gname):
    G = get_graph(gname)
    try:
        color = G.metadata.get_metadata("Color", T.t)
    except KeyError:
        raise MetaNotFound(T.t)
    kwargs = {}
    if T.abundance:
        if T.abundance.operation.t == "<":
            kwargs = {color: lambda e: e.cast("INTEGER").lt(T.abundance.value)}
        if T.abundance.operation.t == "<=":
            kwargs = {color: lambda e: e.cast("INTEGER").le(T.abundance.value)}
        if T.abundance.operation.t == "=":
            kwargs = {color: lambda e: e.cast("INTEGER").eq(T.abundance.value)}
        if T.abundance.operation.t == ">":
            kwargs = {color: lambda e: e.cast("INTEGER").gt(T.abundance.value)}
        if T.abundance.operation.t == ">=":
            kwargs = {color: lambda e: e.cast("INTEGER").ge(T.abundance.value)}

    return G.find_all_nodes(color, kwargs)


@_search.register(query.All)
def _(T, gname):
    G = get_graph(gname)
    return G.find_all_nodes()


@_search.register(query.Partial)
def _(T, gname):
    raise QueryError("Partial is only for Client-side")


@_search.register(query.Degree)
def _(T, gname):
    raise QueryError("Degree is only for Client-side")


@_search.register(query.Selection)
def _(T, gname):
    raise QueryError("Selection is only for Client-side")


@_search.register(query.Loop)
def _(T, gname):
    raise QueryError("Loop is only for Client-side")


@_search.register(query.Kmer)
def _(T, gname):
    G = get_graph(gname)
    if len(T.t) != G.metadata.k:
        raise InvalidKmerSize(f"{T.size} expected {G.metadata.k}")

    kmer = Kmer.from_sequence(DNA(T.kmer)).canonical()
    return search_kmers(gname, (kmer,))


@_search.register(query.Seq)
def _(T, gname):
    G = get_graph(gname)
    if T.smallk:
        small_k = T.smallk.val
        threshold = T.threshold.val if T.threshold else 70
    else:
        # if small_k is not setup we compute one which is valid
        potential_idx = [idx.k for idx in index_info(gname) if idx.k <= len(T.seq)]
        threshold = T.threshold.val if T.threshold else 0
        if not potential_idx:
            if T.threshold:
                raise QueryError("Threshold can't be taken into account without index")
            return G.find_all_nodes(sequence=lambda e: e.like(f"%{T.seq}%"))
        idx = max(potential_idx)
        small_k = None
        if idx != G.metadata.k:
            small_k = idx

    k = small_k if small_k else G.metadata.k
    kmers = set(next(iter(DNA.from_str(T.seq))).enum_canonical_kmer(k))
    iq = search_kmers(gname, kmers, small_k=small_k)
    if threshold:  # if threshold is 0 we don't even take it into account
        nodes = list()
        for node in iq:
            dna = next(DNA.from_str(G.nodes[node]["sequence"]))
            local_kmers = set(dna.enum_canonical_kmer(k))
            if len(local_kmers.intersection(kmers)) / len(local_kmers) >= (
                threshold / 100
            ):
                nodes.append(node)
        if not nodes:
            return empty_iterable_query(G)
        iq = constants_to_iterable_query(G, nodes)
    return iq


@_search.register(query.NodeId)
def _(T, gname):
    G = get_graph(gname)
    return G.nbunch_iter(T.t)


@_search.register(query.And)
def _(T, gname):
    iqs = list(map(lambda t: _search(t, gname), T.t))
    return iqs[0].intersection(*iqs[1:])


@_search.register(query.Or)
def _(T, gname):
    iqs = list(map(lambda t: _search(t, gname), T.t))
    if iqs:
        return iqs[0].union(*iqs[1:])
    G = get_graph(gname)
    return empty_iterable_query(G)


@_search.register(query.Not)
def _(T, gname):
    G = get_graph(gname)
    return G.find_all_nodes().difference(_search(T.t, gname))


@_search.register(query.Parenthesis)
def _(T, gname):
    G = get_graph(gname)
    iq = _search(T.t, gname)
    return G.dialect.helper.IterableQuery(
        G.helper, G.dialect.queries.SelectQuery(iq.query)
    ).map(lambda e: e[0])


def search(name: str, q: str, limit=1000) -> list[int]:
    logger.info(f"Parsing {q}")
    term = parse(q)
    logger.info(f"Query {q} successfully parsed")
    iq = _search(term, name)
    logger.info(f"Query {q} successfully executed")
    if not isinstance(
        iq.query,
        nd.sqlite.queries.SelectQuery,
    ):  # ugly because of nd limit broken on iq with union/intersection
        return list(limit_iterable_query(iq, limit=limit))
    else:
        return list(iq.limit(limit))



import asyncio
import webbrowser

import uvicorn

from vizitig.cli import subparsers
from vizitig.info import clean_log


async def delayed_webrowser_launch(delay: int, args):
    await asyncio.sleep(delay)
    webbrowser.open(f"{args.hostname}:{args.port}", new=2)


async def main(args):
    config = uvicorn.Config(
        "vizitig.api:main_api",
        port=args.port,
        host=args.hostname,
        log_level="info",
    )
    server = uvicorn.Server(config)
    tasks = []
    if not args.no_launch_browser:
        tasks.append(asyncio.create_task(delayed_webrowser_launch(2, args)))
    tasks.append(server.serve())
    await asyncio.gather(*tasks)


def run(args):
    clean_log()
    asyncio.run(main(args))


parser = subparsers.add_parser(
    "run",
    help="run Vizitig locally",
)

parser.add_argument("--port", "-p", help="port", type=int, default=4242)

parser.add_argument("--hostname", "-H", help="host", type=str, default="localhost")
parser.add_argument(
    "--no-launch-browser",
    "-l",
    help="flag to avoid to launch the webbrower",
    action="store_true",
)
parser.set_defaults(func=run)



import gzip
import os
from os.path import isfile, join
from datetime import datetime
from pathlib import Path
from typing import Callable, IO

from vizitig.cli import subparsers
from vizitig.info import get_graph, reload_graph
from vizitig.parsing import parse_fasta_dna, parse_fasta_with_bcalm_abundances
from vizitig.types import Color
from vizitig.color.bulk_update import bulk_annotate_graph
from vizitig.utils import SubLog
from vizitig.errors import FormatInputError, InvalidPath, ParseError
from vizitig.utils import vizitig_logger as logger
from vizitig.types import DNA


now = datetime.now


def main(args):
    if args.abundances:
        color_method = color_with_abundance
    else:
        color_method = color_graph
        # with SubLog("Color"):
        #     color_with_abundance(
        #         args.graph_name,
        #         args.metadata_name,
        #         args.metadata_description,
        #         args.file,
        #         buffer_size=args.buffer_size,
        #     )

    if args.folder and args.metadata_name:
        if args.abundances:
            logger.warn(
                "Abundances cannot be ingested with a single metadata for a folder. Remove metadata name or use vizitig color -h to get help."
            )
        try:
            d = args.description
        except AttributeError:
            d = ""
        with SubLog("Color_with_folder_and_metadata"):
            color_graph_with_folder_and_meta(
                args.graph_name, args.metadata_name, d, args.folder, args.buffer_size
            )

    elif args.folder:
        try:
            assert os.path.isdir(args.folder)
        except AssertionError:
            raise InvalidPath(
                "The provided path is not a folder or is not a valid path. "
            )
        files = [f for f in os.listdir(args.folder) if isfile(join(args.folder, f))]
        for file_name in files:
            file = Path(args.folder) / file_name
            with SubLog("Color"):
                color_method(
                    args.graph_name,
                    file_name.split(".")[0],
                    "",
                    file,
                    buffer_size=args.buffer_size,
                )

    elif args.csv_file:
        graph_name = args.graph_name  # namespace issue
        buffer_size = args.buffer_size  # namespace issue
        try:
            with open(args.csv_file, "r") as handle:
                for line in handle:
                    args = line.split("\t")
                    if len(args) == 2:
                        file_name, metadata_name = args
                        description = ""
                    elif len(args) == 3:
                        file_name, metadata_name, description = args
                    else:
                        raise ParseError(
                            "CSV file does not respect the format. Type vizitig color -h to get documentation."
                        )

                    try:
                        with SubLog("Color"):
                            color_method(
                                graph_name,
                                metadata_name,
                                description,
                                Path(file_name),
                                buffer_size=buffer_size,
                            )
                    except FormatInputError:
                        logger.warning(
                            "File {} it of unknown format. This file will be skipped.".format(
                                file_name
                            )
                        )

        except FileNotFoundError:
            raise InvalidPath("The path to csv file was not found. Aborting")

    else:
        with SubLog("Color"):
            color_method(
                args.graph_name,
                args.metadata_name,
                args.metadata_description,
                args.file,
                args.abundances,
            )


valid_extensions = (".fa", ".fna", ".gz", ".fasta")


def color_graph_with_folder_and_meta(
    graph_name: str, m: str, d: str, folder: Path, buffer_size=10**6
):
    G = get_graph(graph_name)
    color = G.metadata.add_metadata(
        Color(id=m, description=d),
    )
    ret_int = G.metadata.encoder(color)

    def generator():
        files = [f for f in os.listdir(folder) if isfile(join(folder, f))]
        for file_name in files:
            file = Path(folder) / file_name
            if file.suffix not in valid_extensions:
                logger.warn("File extension not supported. Skipping {}".format(file))
            if file.suffix == ".gz":
                handle = gzip.open
            else:
                handle = open
            with handle(file, "rt") as file:
                for dna in parse_fasta_dna(file, G.metadata.k, buffer_size=buffer_size):
                    yield (dna, ret_int)

    bulk_annotate_graph(graph_name, generator)


def color_graph(graph_name, m: str, d: str, file: Path, buffer_size=10**6):
    G = get_graph(graph_name)
    GM = G.metadata
    color = GM.add_metadata(
        Color(id=m, description=d),
    )

    GM.commit_to_graph()

    if file.suffix not in valid_extensions:
        raise FormatInputError(
            f"Invalid format extension {file.suffix} not in {valid_extensions}"
        )
    logger.info(f"Coloring graph {graph_name} with {str(file)}")

    def dna_generator(file=file):
        ret_int = GM.encoder(color)
        openfile = open
        if file.suffix == ".gz":
            openfile = gzip.open
        with openfile(file, "rt") as file:
            for dna in parse_fasta_dna(file, GM.k, buffer_size):
                yield (dna, ret_int)

    with SubLog("bulk_annotate_graph"):
        bulk_annotate_graph(graph_name, dna_generator)
    reload_graph(graph_name)


def color_with_abundance(graph_name, m: str, d: str, file: Path, buffer_size=10**6):
    G = get_graph(graph_name)
    GM = G.metadata
    color = GM.add_metadata(
        Color(id=m, description=d),
    )

    GM.commit_to_graph()

    if file.suffix not in valid_extensions:
        raise FormatInputError(
            f"Invalid format extension {file.suffix} not in {valid_extensions}"
        )
    logger.info(f"Coloring graph {graph_name} with {str(file)}")

    openfile: Callable[..., IO] = open
    if file.suffix == ".gz":
        openfile = gzip.open

    def dna_generator(file=file):
        with openfile(file, "rt") as file:
            for dna, abundance in parse_fasta_with_bcalm_abundances(file):
                assert isinstance(dna, DNA)
                yield (dna, int(abundance))

    with SubLog("bulk_annotate_graph_with_abundances"):
        bulk_annotate_graph(graph_name, dna_generator, color)

    reload_graph(graph_name)


parser = subparsers.add_parser(
    "color",
    help="""Color an existing graph. There are several ways to use this feature : 
    -vizitig color -f file_name -m color_name -> Will color the graph with whatever is in the provided file
    -vizitig color --folder folder_name -> Will color the graph with every file in the provided folder. The name of the color will be the name of the file for each file
    -vizitig color --folder folder_name -m color_name -> Same, but the name of the color will be color_name for all files
    -vizitig color --csv file.csv -> Will use a csv file to color the graph. Csv must respect the following format (tabèseparated):

    /path/to/file1        red           "Sample 1 - control group"
    /path/to/file2        green         "Sample 2 - condition 1"
    /path/to/file3        blue          "Sample 3 - condition 2"
    """,
)

parser.add_argument(
    "--folder",
    metavar="folder",
    type=str,
    help="""
    A folder path. Every parsable file of the folder will be used to color the graph. 
    Use argument -m to give a name to this color, otherwise the name of the files will be used.
    If you want to assign a file to a color and a description, consider using vizitig color --csv. 
    """,
)

parser.add_argument(
    "--csv_file",
    metavar="csv_file",
    type=str,
    help="""A csv file to use for coloring a graph. Must respect the following format (tab-separated):
    /path/to/file1        red           "Sample 1 - control group"
    /path/to/file2        green         "Sample 2 - condition 1"
    /path/to/file3        blue          "Sample 3 - condition 2"
    """,
)
parser.add_argument(
    "graph_name",
    help="A graph name. List possible graph with python3 -m vizitig info",
    metavar="graph",
    type=str,
)

parser.add_argument(
    "-m",
    "--metadata_name",
    help="A key to identify the metadata to add to the graph",
    metavar="metadata_name",
    required=False,
    type=str,
)

parser.add_argument(
    "-d",
    "--metadata-description",
    help="A description of the metadata",
    metavar="description",
    type=str,
    default="",
)

parser.add_argument(
    "-f",
    "--file",
    help="Path toward file containing DNA (fasta or bcalm format)",
    metavar="file",
    type=Path,
)

parser.add_argument(
    "-b",
    "--buffer-size",
    help="Maximum size of a buffer",
    metavar="buffer",
    type=int,
    default=10**6,
)

parser.add_argument(
    "-c",
    "--color",
    help="Default color to use in the vizualisation. Default is None",
    metavar="color",
    type=str,
)

parser.add_argument(
    "-abundances",
    action="store_true",
    default=False,
    help="""Activates abundance injestion. This means that the abundances in the file will be tagged on the graph. File must be a valid BCALM file with defaut abundance formation (i.e 'km:f:value')
    Note that the unitigs of this BCALM may not equals the unitigs of the graph you have currently ingested. In this case, the abundances of unitigs may be found on several nodes. The abundance
    ingestion is not compatible with the 'folder' parameter. You can use csv-file argument if you need to ingest abundances with a lot of files.
    """,
)

parser.set_defaults(func=main)



import functools
import shutil
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import Any, Annotated, cast, Literal
from uuid import uuid4

import networkx as nx
from fastapi import FastAPI, HTTPException, UploadFile, status, Form

from vizitig.env_var import VIZITIG_TMP_DIR
from vizitig import query as vizquery
from vizitig import version as vizversion
from vizitig.api.async_utils import async_subproc
from vizitig.api.errors import (
    NodeNotFoundException,
    NoEmptyGraphNameAllowed,
    NoPathFoundException,
    UnknownFormat,
    UnsupportedExtension,
)
from vizitig.api.path import site
from vizitig.api.types import Alignment
from vizitig.errors import VizitigException
from vizitig.export import export_format, export_graph
from vizitig.generate_graph import generate_graph
from vizitig.index import (
    IndexInfo,
    build_kmer_index as base_build_kmer_index,
    drop_kmer_index as base_drop_kmer_index,
    index_info as base_index_info,
    index_types as base_index_types,
)
from vizitig.info import (
    add_vizitig_graph,
    get_graph,
    reload_graph,
    delete_graph as delete_graph_base,
    graph_info as base_graph_info,
    graphs_list as base_graphs_list,
    rename_graph as rename_graph_base,
)
from vizitig.color import color_graph as base_update_graph
from vizitig.annotate import (
    annotate_genome as base_annotate_genome,
    annotate_other as base_annotate_other,
)
from vizitig.metadata import GraphMetadata, NodeDesc
from vizitig.paths import graph_path_name, log_path_name, root
from vizitig.types import ESign, Metadata

app = FastAPI()


def wrap_error(fct):
    """Wrap the Vizitig Error into HTTP Error"""

    @functools.wraps(fct)
    async def _fct(*args, **kwargs):
        try:
            return await fct(*args, **kwargs)
        except VizitigException as E:
            # This could be more efficient with an appropriate
            # layout of VizitiException.
            raise HTTPException(detail=f"{E!r}", status_code=400)

    return _fct


def get(*args, **kwargs):
    def ndec(f):
        f = wrap_error(f)
        return app.get(*args, operation_id=f.__name__, **kwargs)(f)

    return ndec


def post(*args, **kwargs):
    def ndec(f):
        f = wrap_error(f)
        return app.post(*args, operation_id=f.__name__, **kwargs)(f)

    return ndec


def patch(*args, **kwargs):
    def ndec(f):
        f = wrap_error(f)
        return app.patch(*args, operation_id=f.__name__, **kwargs)(f)

    return ndec


def delete(*args, **kwargs):
    def ndec(f):
        f = wrap_error(f)
        return app.delete(*args, operation_id=f.__name__, **kwargs)(f)

    return ndec


export_path = Path(root, "export")
if not export_path.exists():
    export_path.mkdir()


@get(
    path="/ryalive",
    status_code=status.HTTP_200_OK,
    description="Endpoint to check that server is online",
)
async def ryalive():
    return "yes master !!"


@get(
    path="/version",
    status_code=status.HTTP_200_OK,
    description="Endpoint to get the current vizitig version",
)
async def version() -> str:
    return vizversion


@get(
    path="/load_viz",
    status_code=status.HTTP_200_OK,
    description="Get the list of available visualisation",
)
async def load_viz() -> list[str]:
    return list(map(lambda e: e.stem, site.glob("js/viz/*.js")))


@get(
    path="/list_graphs",
    status_code=status.HTTP_200_OK,
    description="Get the list of available graphs",
)
async def graphs_list() -> dict[str, Any]:
    result = dict()
    for e in base_graphs_list():
        try:
            data = base_graph_info(e, human_size=True)
        except Exception as E:
            data = dict(error=str(E))
        result[e] = data
    return result


@get(
    path="/export_format",
    status_code=status.HTTP_200_OK,
    description="Get the list of available format to export some graph",
)
async def get_export_format() -> list[str]:
    return sorted(export_format)


@delete(
    path="/graphs/{name}",
    status_code=status.HTTP_200_OK,
    description="Delete the graph",
)
async def delete_graph(name: str) -> None:
    delete_graph_base(name)


@get("/log/{name}")
async def get_log(name: str) -> list[str]:
    try:
        with open(log_path_name(name)) as f:
            return list(f)
    except FileNotFoundError:
        return []


@post(
    path="/graph/rename/{old_name}/{new_name}",
    status_code=status.HTTP_200_OK,
    description="Rename the graph",
)
async def rename_graph(old_name, new_name):
    rename_graph_base(old_name, new_name, replace=False)


@post(
    path="/graph/rename",
    status_code=status.HTTP_200_OK,
    description="Rename the graph via a formular",
)
async def rename_graph_form(
    name: Annotated[str, Form()], new_name: Annotated[str, Form()]
):
    rename_graph_base(name, new_name, replace=False)


def fmt_name(name, idx):
    return f"{name}_copy_{idx}"


@get(
    path="/align/{gname}/{nid1}/{nid2}",
    status_code=status.HTTP_200_OK,
    description="Align the unit data of two nodes",
)
async def align(
    gname: str,
    nid1: int,
    nid2: int,
) -> Alignment:
    G = get_graph(gname)
    try:
        seq1 = G.nodes[nid1]["sequence"]
        seq2 = G.nodes[nid2]["sequence"]
    except KeyError as E:
        raise NodeNotFoundException(detail=E.args[0], status_code=404)
    return Alignment.from_seq(seq1, seq2)


@get(
    path="/index/{name}/info",
    status_code=status.HTTP_200_OK,
    description="Get info on the index of the graph",
)
async def index_info(name: str) -> list[IndexInfo]:
    return base_index_info(name)


@get(
    path="/index_types",
    status_code=status.HTTP_200_OK,
    description="Get info on the index of the graph",
)
async def index_types() -> list[str]:
    return base_index_types


@post(
    path="/index/{name}/{index_type}/drop",
    status_code=status.HTTP_200_OK,
    description="Drop indexes of the graph. If the index type is provided, drop only this type",
)
async def drop_index(name: str, index_type: str, small_k: int | None = None):
    if index_type is not None:
        base_drop_kmer_index(name, index_type, small_k=small_k)
    else:
        base_drop_kmer_index(name, index_type, small_k=small_k)


@post(
    path="/index/{name}/build",
    status_code=status.HTTP_200_OK,
    description="Build an index for the graph. If no type is provided, the some index_type is selected somehow",
)
async def build_index(
    name: str, index_type: str | None = None, small_k: int | Literal[""] | None = None
):
    if index_type is None:
        index_type = base_index_types[-1]
    if isinstance(small_k, str):
        small_k = None
    await async_subproc(base_build_kmer_index)(name, index_type, small_k=small_k)


@post(
    path="/graph/color/",
    status_code=status.HTTP_200_OK,
    description="Color the graph with an input file",
)
async def color_graph(
    name: Annotated[str, Form()], file: UploadFile, color: Annotated[str, Form()]
):
    assert file.filename is not None
    path = Path(file.filename)
    tmp = NamedTemporaryFile(prefix=VIZITIG_TMP_DIR, delete=False, suffix=path.suffix)
    shutil.copyfileobj(file.file, tmp)  # type: ignore
    tmp_path = Path(tmp.name)
    await async_subproc(base_update_graph)(name, color, "", tmp_path)
    file.file.close()
    tmp_path.unlink()
    reload_graph(name)


@post(
    path="/graph/annotate_gene/",
    status_code=status.HTTP_200_OK,
    description="Color the graph with an input file",
)
async def annotate_gene(
    name: Annotated[str, Form()], ref_gen: UploadFile, gtf_gff: UploadFile
):
    assert ref_gen.filename
    assert gtf_gff.filename
    path_gen = Path(ref_gen.filename)
    path_gtf_gff = Path(gtf_gff.filename)

    tmp_gen = NamedTemporaryFile(
        prefix=VIZITIG_TMP_DIR, delete=False, suffix=path_gen.suffix
    )
    shutil.copyfileobj(ref_gen.file, tmp_gen)  # type: ignore

    tmp_gtf_gff = NamedTemporaryFile(
        prefix=VIZITIG_TMP_DIR, delete=False, suffix=path_gtf_gff.suffix
    )

    shutil.copyfileobj(gtf_gff.file, tmp_gtf_gff)  # type: ignore

    tmp_path_gen = Path(tmp_gen.name)
    tmp_path_gtf_gff = Path(tmp_gtf_gff.name)

    await async_subproc(base_annotate_genome)(name, tmp_path_gtf_gff, tmp_path_gen)

    ref_gen.file.close()
    gtf_gff.file.close()
    tmp_path_gen.unlink()
    tmp_path_gtf_gff.unlink()
    reload_graph(name)


@post(
    path="/graph/annotate_other/",
    status_code=status.HTTP_200_OK,
    description="Color the graph with an input file",
)
async def annotate_other(
    name: Annotated[str, Form()],
    reference_seqs: UploadFile,
    metadata_type: Annotated[str, Form()],
    gtf_gff: UploadFile | None = None,
):
    assert reference_seqs.filename
    path_reference_seqs = Path(reference_seqs.filename)
    tmp_reference_seqs = NamedTemporaryFile(
        prefix=VIZITIG_TMP_DIR, delete=False, suffix=path_reference_seqs.suffix
    )
    shutil.copyfileobj(reference_seqs.file, tmp_reference_seqs)  # type: ignore
    tmp_ref_seq = Path(tmp_reference_seqs.name)

    if gtf_gff:
        assert gtf_gff.filename
        path_gtf_gff = Path(gtf_gff.filename)
        tmp_gtf_gff = NamedTemporaryFile(
            prefix=VIZITIG_TMP_DIR, delete=False, suffix=path_gtf_gff.suffix
        )

        shutil.copyfileobj(gtf_gff.file, tmp_gtf_gff)  # type: ignore

        tmp_path_gtf_gff = Path(tmp_gtf_gff.name)
    else:
        tmp_path_gtf_gff = None
    await async_subproc(base_annotate_other)(
        name, metadata_type, tmp_ref_seq, tmp_path_gtf_gff
    )

    reference_seqs.file.close()
    tmp_ref_seq.unlink()
    if gtf_gff:
        cast(Path, tmp_path_gtf_gff).unlink()
    reload_graph(name)


@post(
    path="/duplicate/{name}",
    status_code=status.HTTP_200_OK,
    description="Duplicate the graph",
)
async def duplicate(name: str):
    L = set(base_graphs_list())
    path = graph_path_name(name)
    new_name = name
    idx = 0
    while fmt_name(new_name, idx) in L:
        idx += 1

    name = fmt_name(new_name, idx)
    await async_subproc(add_vizitig_graph)(path, name=name, replace=False, copy=True)


@post(
    path="/upload/{name}",
    status_code=status.HTTP_200_OK,
    description="""
    Upload a graph. 
    If replace is set to True, replace existing graph. 
    If check_compatibility is set to true, check is graph is compatible
    with the current vizitig version (default True)""",
)
async def upload_graph(file: UploadFile, name: str):
    assert file.filename is not None
    path = Path(file.filename)
    gname = path.stem
    if name != "":
        gname = name
    if name == "":
        raise NoEmptyGraphNameAllowed(status_code=402)

    if path.suffix in (".fa", ".gz", ".fasta"):
        try:
            tmp = NamedTemporaryFile(
                prefix=VIZITIG_TMP_DIR, delete=False, suffix=path.suffix
            )
            shutil.copyfileobj(file.file, tmp)  # type: ignore
            tmp_path = Path(tmp.name)
        finally:
            file.file.close()
        await async_subproc(generate_graph)(tmp_path, graph_path_name(name), name)
        tmp_path.unlink()
    elif path.suffix == ".db":
        try:
            with NamedTemporaryFile(
                prefix=VIZITIG_TMP_DIR, delete=False, suffix=".db"
            ) as tmp:
                shutil.copyfileobj(file.file, tmp)  # type: ignore
                tmp_path = Path(tmp.name)
        finally:
            file.file.close()
        await async_subproc(add_vizitig_graph)(
            tmp_path,
            name=gname,
            replace=False,
            check_compatibility=True,
        )
    else:
        raise UnsupportedExtension(extension=path.suffix, status_code=400)


@post(
    path="/graphs/{name}/export/{format}",
    status_code=status.HTTP_200_OK,
    description="Return a link to the exported file",
)
async def export_nodes(name: str, format: str, nodes: list[int]) -> str:
    """Exports the current nodes of the graph to a file"""
    # If the format is not supported, raise an error
    if format not in export_format:
        raise UnknownFormat(details=format, status_code=404)

    # Matches the export format of front with end
    # This calls the export_format dict from vizitig/export/__init__.py
    ext = export_format.get(format)

    # Generates a 128 bit random string that will serve as file name
    fname = f"{uuid4()}.{ext}"

    # Build the path
    target_path = Path(export_path, fname)

    await async_subproc(export_graph)(name, nodes, format, target_path)
    return f"export/{fname}"


@get(
    path="/graphs/{name}/info",
    status_code=status.HTTP_200_OK,
    description="Returns informations about the graph",
)
async def graph_info(name: str) -> GraphMetadata:
    return get_graph(name).metadata


def build_node_return(
    G,
    nodes: list[int],
) -> list[tuple[int, NodeDesc]]:
    nodes_with_data = G.subgraph(nodes).nodes(data=True)
    L: list[tuple[int, NodeDesc]] = []
    for x, d in nodes_with_data:
        neighbors: dict[int, ESign] = dict()
        adj = G._adj[x].fold()
        for y, od in adj.items():
            neighbors[y] = od.get("sign", None)
        L.append((x, G.metadata.to_nodedesc(d, neighbors)))
    return L


@get(
    path="/graphs/{name}/parse_query/{query}",
    status_code=status.HTTP_200_OK,
    description="Fetch all nodes with a query",
)
async def parse_query(name: str, query: str) -> vizquery.Term:
    return vizquery.parse(query)


@post(
    path="/graphs/{gname}/filters/{fname}/{filter}",
    status_code=status.HTTP_200_OK,
    description="Add a filter to the graph",
)
async def add_filter(gname: str, fname: str, filter: str):
    G = get_graph(gname)
    G.metadata.add_filter(fname, filter)
    G.metadata.commit_to_graph()


@post(
    path="/graphs/{gname}/filters/{fname}",
    status_code=status.HTTP_200_OK,
    description="Remove a filter from the graph",
)
async def remove_filter(gname: str, fname: str):
    G = get_graph(gname)
    G.metadata.remove_filter(fname)
    G.metadata.commit_to_graph()


@get(
    path="/graphs/{gname}/filters/",
    status_code=status.HTTP_200_OK,
    description="Get all filters of the graph",
)
async def list_filters(gname: str) -> list[tuple[str, str]]:
    G = get_graph(gname)
    return list(G.metadata.get_filters())


@get(
    path="/graphs/{gname}/states/",
    status_code=status.HTTP_200_OK,
    description="Get all the states of the graph",
)
async def list_states(gname: str) -> list[str]:
    G = get_graph(gname)
    return list(G.metadata.get_states())


@get(
    path="/graphs/{gname}/state/{state_name}",
    status_code=status.HTTP_200_OK,
    description="Get the state of the graph by name",
)
async def get_state_by_name(gname: str, state_name: str) -> str:
    G = get_graph(gname)
    return G.metadata.get_state(state_name)


@post(
    path="/graphs/{gname}/state/{state_name}/{state}",
    status_code=status.HTTP_200_OK,
    description="Set the state of the graph by name",
)
async def save_state(gname: str, state_name: str, state: str):
    G = get_graph(gname)
    G.metadata.set_state(state_name, state)


@post(
    path="/graphs/{gname}/delete_state/{state_name}",
    status_code=status.HTTP_200_OK,
    description="Delete the state of the graph by name",
)
async def delete_state(gname: str, state_name: str):
    G = get_graph(gname)
    G.metadata.delete_state(state_name)


@get(
    path="/graphs/{name}/find/query/{query}",
    status_code=status.HTTP_200_OK,
    description="Fetch all nodes with a query",
)
async def find_with_query(name: str, query: str) -> list[tuple[int, NodeDesc]]:
    G = get_graph(name)
    # nodes = vizquery.search(name, query)
    nodes = await async_subproc(vizquery.search)(name, query)
    return build_node_return(G, list(nodes))


@post(
    path="/graphs/{name}/node_data",
    status_code=status.HTTP_200_OK,
    description="Get nodes data in argument (POST only)",
)
async def nodes_data(name: str, nodes: list[int]) -> list[tuple[int, NodeDesc]]:
    G = get_graph(name)
    return build_node_return(G, nodes)


@get(
    path="/graphs/{name}/{nid}",
    status_code=status.HTTP_200_OK,
    description="Get data from `nid`",
)
async def all_nid(name: str, nid: int) -> NodeDesc:
    G = get_graph(name)
    try:
        d = G.nodes[nid].fold()
        nd = G.metadata.to_nodedesc(d, [y for y in G[nid]])
        return nd
    except KeyError:
        raise NodeNotFoundException(detail=nid, status_code=404)


@get(
    path="/graphs/{name}/meta/types",
    status_code=status.HTTP_200_OK,
    description="Return the metadata information stored in the graph",
)
async def get_all_types(name: str) -> list[str]:
    G = get_graph(name)
    return G.metadata.types_list


@get(
    path="/graphs/{gname}/meta/types/{type}",
    status_code=status.HTTP_200_OK,
    description="Return the metadata information stored in the graph",
)
async def all_metadata_of_types(gname: str, type: str) -> list[Metadata]:
    G = get_graph(gname)
    return G.metadata.all_metadata_by_types(type)


@get(
    path="/graphs/{name}/path/{source}/{target}",
    status_code=status.HTTP_200_OK,
    description="Get the path if it exists between `source` and `target`.",
)
async def get_path(name: str, source: int, target: int):
    G = get_graph(name)
    try:
        return nx.shortest_path(G, source, target)
    except nx.exception.NetworkXNoPath:
        return NoPathFoundException(detail=(source, target), status_code=404)
    except nx.exception.NodeNotFound:
        return NodeNotFoundException(detail=(source, target), status_code=404)



from typing_extensions import Self

from pydantic import BaseModel

from vizitig.types import DNAPython, KmerPython


class Alignment(BaseModel):
    score: int | float
    align_seq1: list[str | None]
    align_seq2: list[str | None]

    @classmethod
    def from_seq(cls, s1: str, s2: str) -> Self:
        aligner = cls.alignment
        alignement1 = aligner(  # type: ignore
            s1,
            s2,
        )
        alignement2 = aligner(  # type: ignore
            str(
                KmerPython.from_sequence(
                    next(DNAPython.from_str(s1))
                ).reverse_complement()
            ),
            s2,
        )
        alignement = max(alignement1, alignement2, key=lambda e: e[2])

        return cls(
            score=alignement[2],
            align_seq1=[e for e in alignement[0]],
            align_seq2=[e for e in alignement[1]],
        )

    @classmethod
    def alignment(
        cls,
        seq1,
        seq2,
        match_score=10,
        mismatch_score=-100,
        gap_open=-100,
        gap_extend=0,
    ):
        n, m = len(seq1), len(seq2)

        # Use native Python lists instead of NumPy arrays
        score_matrix = [[0] * (m + 1) for _ in range(n + 1)]
        traceback = [[0] * (m + 1) for _ in range(n + 1)]

        for i in range(1, n + 1):
            score_matrix[i][0] = 0
            traceback[i][0] = 2

        for j in range(1, m + 1):
            score_matrix[0][j] = 0
            traceback[0][j] = 1

        for i in range(1, n + 1):
            for j in range(1, m + 1):
                match = score_matrix[i - 1][j - 1] + (
                    match_score if seq1[i - 1] == seq2[j - 1] else mismatch_score
                )
                delete = score_matrix[i - 1][j] + (
                    gap_extend if traceback[i - 1][j] == 2 else gap_open
                )
                insert = score_matrix[i][j - 1] + (
                    gap_extend if traceback[i][j - 1] == 1 else gap_open
                )

                score_matrix[i][j] = max(match, delete, insert)

                if score_matrix[i][j] == match:
                    traceback[i][j] = 0
                elif score_matrix[i][j] == insert:
                    traceback[i][j] = 1
                else:
                    traceback[i][j] = 2

        max_score = max(max(score_matrix[n]), max(row[m] for row in score_matrix))
        if max(score_matrix[n]) >= max(row[m] for row in score_matrix):
            i, j = n, score_matrix[n].index(max(score_matrix[n]))
        else:
            i, j = max((row[m], idx) for idx, row in enumerate(score_matrix))[1], m

        aligned_seq1, aligned_seq2 = "", ""

        while i > 0 or j > 0:
            if i == 0:
                aligned_seq1 = "-" + aligned_seq1
                aligned_seq2 = seq2[j - 1] + aligned_seq2
                j -= 1
            elif j == 0:
                aligned_seq1 = seq1[i - 1] + aligned_seq1
                aligned_seq2 = "-" + aligned_seq2
                i -= 1
            elif traceback[i][j] == 0:
                aligned_seq1 = seq1[i - 1] + aligned_seq1
                aligned_seq2 = seq2[j - 1] + aligned_seq2
                i -= 1
                j -= 1
            elif traceback[i][j] == 1:
                aligned_seq1 = "-" + aligned_seq1
                aligned_seq2 = seq2[j - 1] + aligned_seq2
                j -= 1
            else:
                aligned_seq1 = seq1[i - 1] + aligned_seq1
                aligned_seq2 = "-" + aligned_seq2
                i -= 1

        # Ensure full sequences are returned
        aligned_seq1, aligned_seq2 = cls.restore_full_sequences(
            seq1, seq2, aligned_seq1, aligned_seq2
        )

        return aligned_seq1, aligned_seq2, max_score

    @staticmethod
    def restore_full_sequences(seq1, seq2, aligned_seq1, aligned_seq2):
        start1 = seq1.find(aligned_seq1.replace("-", ""))
        start2 = seq2.find(aligned_seq2.replace("-", ""))

        aligned_seq1 = seq1[:start1] + aligned_seq1
        aligned_seq2 = seq2[:start2] + aligned_seq2

        aligned_seq1 += seq1[start1 + len(aligned_seq1.replace("-", "")) :]
        aligned_seq2 += seq2[start2 + len(aligned_seq2.replace("-", "")) :]

        if len(aligned_seq1) < len(aligned_seq2):
            aligned_seq1 += "-" * (len(aligned_seq2) - len(aligned_seq1))
        if len(aligned_seq2) < len(aligned_seq1):
            aligned_seq2 += "-" * (len(aligned_seq1) - len(aligned_seq2))

        return aligned_seq1, aligned_seq2



from itertools import groupby
from typing import Tuple, Callable, Iterator, MutableMapping, List, Any, cast
from statistics import median

from vizitig.index import load_kmer_index
from vizitig.info import get_graph
from vizitig.index import temporary_kmerindex
from vizitig.types import DNA, SubseqData, Color
from vizitig.env_var import VIZITIG_NO_TMP_INDEX
from vizitig.utils import vizitig_logger as logger, progress, IteratableFromGenerator


def bulk_annotate_graph(
    gname: str,
    generator: Callable[[], Iterator[Tuple[DNA, int]]],
    color: Color | None = None,
    agg: Callable = median,
    metas: list[SubseqData] | None = None,
):
    Graph = get_graph(gname)
    k = Graph.metadata.k
    index = load_kmer_index(gname)

    def pos_int_generator():
        for dna, i in generator():
            yield (dna, abs(i))

    if VIZITIG_NO_TMP_INDEX:
        it_dna = pos_int_generator()
        it_kmer = (
            (kmer, key) for dna, key in it_dna for kmer in dna.enum_canonical_kmer(k)
        )
        res_join = groupby(sorted(index.join(it_kmer)), key=lambda e: e[0])
    else:
        tmp_index = temporary_kmerindex(
            pos_int_generator,
            k,
            index_type=index.index_type.__name__,
            shard_number=index.shard_number,
        )
        res_join = groupby(sorted(index.join_index(tmp_index)), key=lambda e: e[0])
        # sorted could probably be removed in index join is in sorted order.

    # This could be done by chunk, if too big to hold in RAM
    graph_annotations: List[
        Tuple[Any, MutableMapping[SubseqData, int] | MutableMapping[Color, int]]
    ] = []

    annotation_size = 0

    if color:
        for inode, f in res_join:
            graph_annotations.append(
                (inode, {color: agg(tuple(abundance for _, abundance in f))})
            )
            annotation_size += 1
    else:
        if metas is not None:
            decoder_dict: dict[int, SubseqData] = {
                m.offset: m for m in metas if m.offset is not None
            }

            assert len(decoder_dict) == len(metas)

            def decoder(e: int) -> SubseqData:
                return decoder_dict[e]
        else:

            def decoder(e: int) -> SubseqData:
                return Graph.metadata.decoder(-e)

        meta_to_add: set[SubseqData] = set()
        for inode, E in res_join:
            d: MutableMapping[SubseqData, int] = {}
            for _, meta_int in E:
                meta = decoder(meta_int)
                d[meta] = -1
                if not isinstance(meta, Color) and meta.gene:
                    upstream_gene = SubseqData(
                        id=meta.gene,
                        type="Gene",
                        start=meta.start,
                        stop=meta.stop,
                        list_attr=meta.list_attr,
                    )
                    meta_to_add.add(upstream_gene)
                    d[upstream_gene] = -1

                if not isinstance(meta, Color) and meta.transcript:
                    upstream_transcript = SubseqData(
                        id=meta.transcript,
                        type="Transcript",
                        start=meta.start,
                        stop=meta.stop,
                        list_attr=meta.list_attr,
                    )
                    meta_to_add.add(upstream_transcript)
                    d[upstream_transcript] = -1
                    if upstream_transcript.gene:
                        upstream_gene = SubseqData(
                            id=upstream_transcript.gene,
                            type="Gene",
                            start=meta.start,
                            stop=meta.stop,
                            list_attr=meta.list_attr,
                        )
                        meta_to_add.add(upstream_gene)
                        d[upstream_gene] = -1

            graph_annotations.append((inode, d))
            annotation_size += len(d)
        if meta_to_add:
            meta_to_add = set(Graph.metadata.add_iterative_metadatas(meta_to_add))
            meta_to_add_dict = {m.short_repr(): m for m in meta_to_add}
            for i in range(len(graph_annotations)):
                inode, val = graph_annotations[i]
                new_d = {
                    meta_to_add_dict.get(m.short_repr(), cast(SubseqData, m)): v
                    for m, v in val.items()
                }
                graph_annotations[i] = (inode, new_d)

    logger.info(f"Adding {annotation_size} annotations to graph")
    iter_from_gene = IteratableFromGenerator(
        lambda: progress(graph_annotations, total=len(graph_annotations))
    )
    Graph.add_node_data_from(iter_from_gene)



import shutil
from abc import abstractmethod
from collections.abc import Mapping, Set
from pathlib import Path
from typing import Iterable, Tuple, Type, Iterator, Callable
from typing_extensions import Self

from pydantic import BaseModel
from tempfile import TemporaryDirectory, NamedTemporaryFile

from vizitig.errors import (
    InvalidIndexPath,
    InvalidShardPath,
    MissingShardPath,
    IndexNotReady,
)
from vizitig.info import get_graph, GraphLogger
from vizitig.paths import index_path_name
from vizitig.types import Kmer, DNA
from vizitig.env_var import (
    VIZITIG_TMP_DIR,
    VIZITIG_NO_PARALLEL_INDEX,
    VIZITIG_PROC_NUMBER,
)
from vizitig.utils import sizeof_fmt
from vizitig.utils import vizitig_logger as logger, SubLog
from vizitig.utils import progress
from multiprocessing import Process


class IndexInfo(BaseModel):
    gname: str
    type: str
    size: int
    k: int

    def __repr__(self):
        return f"Index({self.type}, {sizeof_fmt(self.size)}, k={self.k})"


class BaseShard:
    def __init__(self, path: Path, shard_idx: int, shard_number: int, k: int):
        self._path = path
        self._shard_idx = shard_idx
        self._shard_number = shard_number
        self._k = k

    @property
    def path(self):
        return self._path

    @property
    def k(self):
        return self._k

    def get_temp_file(self) -> Path:
        file = NamedTemporaryFile(prefix=VIZITIG_TMP_DIR)
        file.close()
        return Path(file.name)


class SetShard(BaseShard, Set[Kmer]):
    priority: int = 0
    subclasses: dict[str, Type["SetShard"]] = dict()

    @classmethod
    def __init_subclass__(cls):
        SetShard.subclasses[cls.__name__] = cls

    @classmethod
    @abstractmethod
    def build_kmer(
        cls,
        path: Path,
        shard_idx: int,
        shard_number: int,
        kmer_iter: Iterator[Kmer],
        k: int,
    ) -> Self:
        """Build the index from the Kmer.
        Implementation of this method must filter out kmer_iter them self
        """
        ...

    @classmethod
    def build_dna(
        cls,
        path: Path,
        shard_idx: int,
        shard_number: int,
        dna_iter: Iterator[DNA],
        k: int,
    ) -> Self:
        kmer_iter = (kmer for dna in dna_iter for kmer in dna.enum_canonical_kmer(k))
        return cls.build_kmer(path, shard_idx, shard_number, kmer_iter, k)


class Shard(BaseShard, Mapping[Kmer, int]):
    priority: int = 0
    subclasses: dict[str, Type["Shard"]] = dict()
    Set: Type[SetShard]

    @classmethod
    def __init_subclass__(cls):
        Shard.subclasses[cls.__name__] = cls

    @classmethod
    @abstractmethod
    def build_kmer(
        cls,
        path: Path,
        shard_idx: int,
        shard_number: int,
        kmer_iter: Iterator[Tuple[Kmer, int]],
        k: int,
    ) -> Self:
        """Build the index from the Kmer.
        Implementation of this method must filter out kmer_iter them self
        """
        ...

    @classmethod
    def build_dna(
        cls,
        path: Path,
        shard_idx: int,
        shard_number: int,
        dna_iter: Iterator[Tuple[DNA, int]],
        k: int,
    ) -> Self:
        kmer_iter = (
            (kmer, val) for dna, val in dna_iter for kmer in dna.enum_canonical_kmer(k)
        )
        return cls.build_kmer(path, shard_idx, shard_number, kmer_iter, k)

    def join(self, kmer_iter: Iterator[Tuple[Kmer, int]]) -> Iterable[Tuple[int, int]]:
        for kmer, val in kmer_iter:
            if kmer in self:
                for nid in self.get_all(kmer):
                    yield (nid, val)

    def join_index(self, other: Self) -> Iterable[Tuple[int, int]]:
        return self.join(iter(other.items()))

    def intersection(self, kmer_iter: Iterator[Kmer]) -> Iterable[int]:
        for kmer in kmer_iter:
            if kmer in self:
                yield from self.get_all(kmer)

    def intersection_index(self, other) -> Iterable[int]:
        return self.intersection(iter(other))

    @abstractmethod
    def get_all(self, kmer: Kmer) -> Iterable[int]: ...


class BaseKmerIndex:
    def __init__(self, path: Path, index_type: Type[Shard | SetShard], k: int):
        self._index_type = index_type
        self._path = path
        self._k = k
        self.check_path()

    @property
    def index_type(self) -> Type[Shard | SetShard]:
        return self._index_type

    @property
    def k(self):
        return self._k

    def check_path(self):
        if not self.path.is_dir():
            raise InvalidIndexPath(f"{self.path} is not a directory")
        shard_names = set()
        for p in self.path.glob("*"):
            if not p.name.isnumeric():
                raise InvalidShardPath(f"{p} is not a valid shard path")
            shard_names.add(int(p.name))
        if not shard_names:
            raise IndexNotReady()
        expected_shard_name = set(range(max(shard_names) + 1))
        missing_path = expected_shard_name - shard_names
        if missing_path:
            logger.warning("A shard path is missing : {}".format(missing_path))
            #raise MissingShardPath(missing_path)
        self._shard_nb = max(shard_names) + 1

    @property
    def shard_number(self) -> int:
        return self._shard_nb

    @property
    def shards(self) -> Tuple[Shard | SetShard, ...]:
        if not hasattr(self, "_shards"):
            self._shards = tuple(
                self.index_type(self.path / str(i), i, self.shard_number, self.k)
                for i in range(self.shard_number)
            )
        return self._shards

    @property
    def file_size(self) -> int:
        if self.path.is_dir():
            return sum(f.stat().st_size for f in self.path.glob("**/*") if f.is_file())
        return self.path.stat().st_size

    def __iter__(self) -> Iterator[Kmer]:
        for shard in self.shards:
            yield from shard

    def __len__(self) -> int:
        return sum(map(len, self.shards))

    @property
    def path(self) -> Path:
        return self._path

    def drop(self):
        if self.path.is_dir():
            shutil.rmtree(self.path)
        else:
            self.path.unlink()


class KmerSetIndex(BaseKmerIndex, Set[Kmer]):
    @classmethod
    def _build_kmer(
        cls,
        path: Path,
        index_type: Type[SetShard],
        kmer_iter: Callable[[], Iterator[Kmer]],
        shard_number: int,
        k: int,
    ) -> Self:
        if path.exists():
            logger.warning(f"Index {path}, erasing and rebuilding")
            shutil.rmtree(path)
            path.mkdir()

        process = []
        for shard_index in range(shard_number):
            shard_path = path / str(shard_index)
            if VIZITIG_NO_PARALLEL_INDEX:
                index_type.build_kmer(
                    shard_path, shard_index, shard_number, kmer_iter(), k
                )
            else:
                proc = Process(
                    target=index_type.build_kmer,
                    args=(shard_path, shard_index, shard_number, kmer_iter(), k),
                )
                proc.start()
                process.append(proc)
                if (shard_index + 1) % VIZITIG_PROC_NUMBER == 0:
                    for proc in process:
                        proc.join()
                    process = []

        if not VIZITIG_NO_PARALLEL_INDEX:
            for proc in process:
                proc.join()
        return cls(path, index_type, k)

    @classmethod
    def _build_dna(
        cls,
        path: Path,
        index_type: Type[SetShard],
        dna_iter: Callable[[], Iterator[DNA]],
        shard_number: int,
        k: int,
    ) -> Self:
        if path.exists():
            logger.warning(f"Index {path}, erasing and rebuilding")
            shutil.rmtree(path)
        path.mkdir()

        process = []
        for shard_index in range(shard_number):
            shard_path = path / str(shard_index)
            if VIZITIG_NO_PARALLEL_INDEX:
                index_type.build_dna(
                    shard_path, shard_index, shard_number, dna_iter(), k
                )
            else:
                proc = Process(
                    target=index_type.build_dna,
                    args=(shard_path, shard_index, shard_number, dna_iter(), k),
                )
                proc.start()
                process.append(proc)
                if (shard_index + 1) % VIZITIG_PROC_NUMBER == 0:
                    for proc in process:
                        proc.join()
                    process = []

        if not VIZITIG_NO_PARALLEL_INDEX:
            for proc in process:
                proc.join()

        return cls(path, index_type, k)

    def __contains__(self, kmer: Kmer) -> bool:
        n: int = self.shard_number
        i = hash(kmer) % n
        return kmer in self.shards[i]


class KmerIndex(BaseKmerIndex, Mapping[Kmer, int]):
    @classmethod
    def _build_kmer(
        cls,
        path: Path,
        index_type: Type[Shard],
        kmer_iter: Callable[[], Iterator[Tuple[Kmer, int]]],
        shard_number: int,
        k: int,
    ) -> Self:
        if path.exists():
            logger.warning(f"Index {path}, erasing and rebuilding")
            shutil.rmtree(path)
            path.mkdir()

        process = []
        for shard_index in range(shard_number):
            shard_path = path / str(shard_index)
            if VIZITIG_NO_PARALLEL_INDEX:
                index_type.build_kmer(
                    shard_path, shard_index, shard_number, kmer_iter(), k
                )
            else:
                proc = Process(
                    target=index_type.build_kmer,
                    args=(shard_path, shard_index, shard_number, kmer_iter(), k),
                )
                proc.start()
                process.append(proc)
                if (shard_index + 1) % VIZITIG_PROC_NUMBER == 0:
                    for proc in process:
                        proc.join()
                    process = []

        if not VIZITIG_NO_PARALLEL_INDEX:
            for proc in process:
                proc.join()
        return cls(path, index_type, k)

    @classmethod
    def _build_dna(
        cls,
        path: Path,
        index_type: Type[Shard],
        dna_iter: Callable[[], Iterator[Tuple[DNA, int]]],
        shard_number: int,
        k: int,
    ) -> Self:
        if path.exists():
            logger.warning(f"Index {path}, erasing and rebuilding")
            shutil.rmtree(path)
        path.mkdir()

        process = []
        for shard_index in range(shard_number):
            shard_path = path / str(shard_index)
            if VIZITIG_NO_PARALLEL_INDEX:
                index_type.build_dna(
                    shard_path, shard_index, shard_number, dna_iter(), k
                )
            else:
                proc = Process(
                    target=index_type.build_dna,
                    args=(shard_path, shard_index, shard_number, dna_iter(), k),
                )
                proc.start()
                process.append(proc)
                if (shard_index + 1) % VIZITIG_PROC_NUMBER == 0:
                    for proc in process:
                        proc.join()
                    process = []

        if not VIZITIG_NO_PARALLEL_INDEX:
            for proc in process:
                proc.join()

        return cls(path, index_type, k)

    def join(self, kmer_iter: Iterator[Tuple[Kmer, int]]) -> Iterable[Tuple[int, int]]:
        for kmer, val in kmer_iter:
            if kmer in self:
                for nid in self.get_all(kmer):
                    yield (nid, val)

    def join_index(self, other: "KmerIndex") -> Iterable[Tuple[int, int]]:
        assert self.k == other.k
        if (
            self.index_type is other.index_type
            and other.shard_number == self.shard_number
        ):
            for left, right in zip(self.shards, other.shards):
                assert isinstance(left, Shard)
                assert isinstance(right, Shard)
                try:
                    yield from left.join_index(right)
                except FileNotFoundError:
                    logger.warning("A file was not found. Aborting this part of the join.")
        else:
            yield from self.join(iter(other.items()))

    def intersection(self, kmer_iter: Iterator[Kmer]) -> Iterable[int]:
        for kmer in kmer_iter:
            if kmer in self:
                yield from self.get_all(kmer)

    def intersection_index(self, other: KmerSetIndex) -> Iterable[int]:
        with SubLog("intersection_index"):
            assert issubclass(self.index_type, Shard)
            if (
                self.index_type.Set is other.index_type
                and other.shard_number == self.shard_number
            ):
                for i, (left, right) in enumerate(zip(self.shards, other.shards)):
                    logger.info(f"Shard: {i}")
                    assert isinstance(left, Shard)
                    assert isinstance(right, SetShard)
                    yield from left.intersection_index(right)
            else:
                logger.warning(
                    "Shard or IndexType mismatch: fall back to (slow) iterator intersection"
                )
                yield from self.intersection(iter(other))

    def __getitem__(self, kmer: Kmer) -> int:
        n: int = self.shard_number
        i = hash(kmer) % n
        shard = self.shards[i]
        assert isinstance(shard, Shard)
        return shard[kmer]

    def get_all(self, kmer: Kmer) -> Iterable[int]:
        n: int = self.shard_number
        i = hash(kmer) % n
        shard = self.shards[i]
        assert isinstance(shard, Shard)
        return shard.get_all(kmer)


def graph_index_path(gname: str, index_type: Type[Shard]):
    return index_path_name(gname) / index_type.__name__


def smallk_graph_index_path(gname: str, small_k, index_type: Type[Shard]):
    (index_path_name(gname) / "small_k" / str(small_k)).mkdir(
        exist_ok=True, parents=True
    )
    return index_path_name(gname) / "small_k" / str(small_k) / index_type.__name__


class GraphIndex(KmerIndex):
    _gname: str | None = None

    def info(self) -> IndexInfo:
        assert self._gname is not None
        return IndexInfo(
            gname=self._gname,
            type=self._index_type.__name__,
            size=self.file_size,
            k=self.k,
        )

    @classmethod
    def from_graph(
        cls, gname: str, index_type: Type[Shard], small_k: int | None = None
    ) -> Self:
        G = get_graph(gname)
        if small_k and small_k != G.metadata.k:
            idx = cls(
                smallk_graph_index_path(gname, small_k, index_type), index_type, small_k
            )
        else:
            idx = cls(graph_index_path(gname, index_type), index_type, G.metadata.k)
        idx._gname = gname
        return idx

    @property
    def gname(self):
        return self._gname

    @classmethod
    def build_kmer(
        cls,
        gname: str,
        index_type: Type[Shard],
        shard_number: int,
        small_k: int | None = None,
    ) -> Self:
        G = get_graph(gname)
        if not small_k:
            k = G.metadata.k
        else:
            k = small_k

        def kmer_iter():
            it = G.nbunch_iter(data="sequence")
            index_iter = (
                (kmer, nid)
                for nid, seq in progress(it, total=G.metadata.size)
                for kmer in DNA(seq).enum_canonical_kmer(k)
            )
            return index_iter

        if small_k:
            path = smallk_graph_index_path(gname, small_k, index_type)
        else:
            path = graph_index_path(gname, index_type)

        with GraphLogger(
            gname, f"Index ({index_type.__name__}) Build ({path}) from kmer"
        ):
            return cls._build_kmer(
                path,
                index_type,
                kmer_iter,
                shard_number,
                k,
            )

    @classmethod
    def build_dna(
        cls,
        gname: str,
        index_type: Type[Shard],
        shard_number: int,
        small_k: int | None = None,
    ):
        G = get_graph(gname)
        if small_k:
            k = small_k
            path = smallk_graph_index_path(gname, small_k, index_type)
        else:
            k = G.metadata.k
            path = graph_index_path(gname, index_type)

        def dna_iter():
            it = G.nbunch_iter(data="sequence")
            return map(lambda e: (DNA(e[1]), e[0]), it)

        with GraphLogger(
            gname, f"Index ({index_type.__name__}) Build ({path}) from DNA"
        ):
            return cls._build_dna(
                path,
                index_type,
                dna_iter,
                shard_number,
                k,
            )


class TemporaryKmerSet(KmerSetIndex):
    _d: TemporaryDirectory | None = None

    @classmethod
    def build_kmer(
        cls,
        index_type: Type[Shard],
        kmer_iter: Callable[[], Iterator[Kmer]],
        shard_number: int,
        k: int,
    ) -> Self:
        d = TemporaryDirectory(prefix=VIZITIG_TMP_DIR)
        assert index_type.Set is not None
        idx = cls._build_kmer(Path(d.name), index_type.Set, kmer_iter, shard_number, k)
        idx._d = d
        return idx

    @classmethod
    def build_dna(
        cls,
        index_type: Type[Shard],
        dna_iter: Callable[[], Iterator[DNA]],
        shard_number: int,
        k: int,
    ) -> Self:
        d = TemporaryDirectory(prefix=VIZITIG_TMP_DIR)
        assert index_type.Set is not None
        idx = cls._build_dna(Path(d.name), index_type.Set, dna_iter, shard_number, k)
        idx._d = d
        return idx

    def __del__(self):
        self._d.cleanup()


class TemporaryKmerIndex(KmerIndex):
    _d: TemporaryDirectory | None = None

    @classmethod
    def build_kmer(
        cls,
        index_type: Type[Shard],
        kmer_iter: Callable[[], Iterator[Tuple[Kmer, int]]],
        shard_number: int,
        k: int,
    ) -> Self:
        d = TemporaryDirectory(prefix=VIZITIG_TMP_DIR)
        idx = cls._build_kmer(Path(d.name), index_type, kmer_iter, shard_number, k)
        idx._d = d
        return idx

    @classmethod
    def build_dna(
        cls,
        index_type,
        dna_iter: Callable[[], Iterator[Tuple[DNA, int]]],
        shard_number: int,
        k: int,
    ) -> Self:

        d = TemporaryDirectory(prefix=VIZITIG_TMP_DIR)
        idx = cls._build_dna(Path(d.name), index_type, dna_iter, shard_number, k)
        idx._d = d
        return idx

    def __del__(self):
        self._d.cleanup()



class VizitigException(Exception): ...


class NoGraphException(VizitigException): ...


class GraphExists(VizitigException): ...


class UnknownIndexType(VizitigException): ...


class FormatInputError(VizitigException): ...


class VizIndexError(VizitigException): ...


class NoData(VizIndexError): ...


class IndexNotReady(VizIndexError): ...


class NoIndex(VizIndexError): ...


class KmerIndexNeeded(VizIndexError): ...


class IndexAlreadyExists(VizIndexError): ...


class InvalidIndexPath(VizIndexError): ...


class InvalidShardPath(VizIndexError): ...


class MissingShardPath(VizIndexError): ...


class IncompatibleViziGraph(VizitigException): ...


class NotAViziGraphError(VizitigException): ...


class FileExists(VizitigException): ...


class NameAlreadyTaken(VizitigException): ...


class CannotRetrieveGraphError(VizitigException): ...


class EmptyExportError(VizitigException): ...


class QueryError(VizitigException): ...


class ParseError(QueryError): ...


class KmerParsingError(ParseError): ...


class MetadataError(VizitigException): ...


class MetadataDuplicate(MetadataError): ...


class MetaNotFound(MetadataError): ...


class InvalidKmerSize(QueryError): ...


class GeneNotFound(QueryError): ...


class TranscriptNotFound(QueryError): ...


class MetaDataNotFound(QueryError): ...


class FixtureError(VizitigException): ...


class InvalidPath(VizitigException): ...


class AbundanceNotFound(VizitigException): ...



import gzip
import logging
import re
from itertools import groupby
from pathlib import Path
from typing import (
    Iterable,
    Iterator,
    List,
    NamedTuple,
    TextIO,
    Tuple,
    IO,
)

from vizitig.metadata import GraphMetadata
from vizitig.types import DNA, ESign, SubseqData
from vizitig.utils import vizitig_logger
from vizitig.errors import FormatInputError


class NodeHeader(NamedTuple):
    node_id: int
    metadatas: List[SubseqData]
    sequence: DNA
    successors: tuple[tuple[int, ESign], ...]
    occurence: float | None = None


dna_parse = re.compile(r"(([ACGTacgt]+\n)*)", re.DOTALL)


def parse_fasta_with_bcalm_abundances(f: IO) -> Iterator[Tuple[DNA, float]]:
    buffer = str()
    for i, line in enumerate(f):
        if line[0] == ">":
            if "km:f:" in line:
                if i > 0:
                    for elem in DNA.from_str(buffer):
                        yield (elem, abundance)  # noqa: F821
                    buffer = str()
                abundance: float = float(re.findall(r"k[am]:f:(\d+\.\d+)", line)[0])
                assert isinstance(abundance, float)
        else:
            buffer += line.strip().upper()
    for elem in DNA.from_str(buffer):
        yield (elem, abundance)


def parse_fasta_dna(f: TextIO, k: int, buffer_size: int = 10**8) -> Iterator[DNA]:
    buffer = ""
    for line in f:
        if line[0] == ">":
            yield from DNA.from_str(buffer)
            buffer = ""
        else:
            buffer += line.strip().upper()
    yield from DNA.from_str(buffer)


# def parse_fasta(f: TextIO, k: int, buffer_size: int = 10**6) -> Iterator[Kmer]:
#     for dna in parse_fasta_dna(f, k, buffer_size=buffer_size):
#         yield from dna.enum_canonical_kmer(k)


def parse_reference_sequences_dna(
    file_path: Path, GM: GraphMetadata, annotation_data: dict[str, SubseqData]
) -> Iterable[Tuple[SubseqData, DNA]]:
    for bulk in parse_transcript_or_exon_refseq(file_path):
        assert bulk[0] == ">"
        name, remain = bulk.split(maxsplit=1)
        parse_name = annotation_data.get(name[1:])
        if not parse_name:
            continue
        yield from (
            (parse_name, dna) for dna in DNA.from_str(remain.upper().replace("\n", ""))
        )


def parse_transcript_or_exon_refseq(file_path: Path) -> Iterator[str]:
    """
    Amazing and simple parser
    """
    with open(file_path) as file:
        buffer = ""
        for line in file:
            if line[0] == ">" and buffer:
                yield buffer
                buffer = line
            else:
                buffer += line
        yield buffer


def parse_genes(gene_desc: str) -> Iterator[SubseqData]:
    for gene in gene_desc.split(";"):
        gene_name, transcripts_desc = gene.split(":", maxsplit=1)
        transcripts = list(transcripts_desc.split(","))
        yield SubseqData(
            id=gene_name,
            type="Gene",
            list_attr=list(),
            start=-1,
            stop=-1,
        )  # We don't know yet start and stop. TODO
        for t in transcripts:
            yield SubseqData(
                id=t,
                type="Transcript",
                list_attr=list(),
                start=-1,
                stop=-1,
            )


find_succ = re.compile(r"L:([+-]):(\d+):([+-])")
find_annotation = re.compile(r"genes:\[(.*?)\]")
occurence_pattern = re.compile(r"(?<=km:f:)[+-]?\d+(?:\.\d+)?")


def parse_one(line: str, seq: str) -> NodeHeader:
    """Function parsing one line of BCALM file
    Its important to note that one line refers
    to the one line that was grouped by the previous
    function
    Hereby one line is the header of a bcalm graph
    plus its sequence
    """
    assert line[0] == ">"
    gene_annotations = find_annotation.findall(line)
    assert len(gene_annotations) <= 1
    if gene_annotations:
        parsed_gene_annotations = list(parse_genes(gene_annotations[0]))
    else:
        parsed_gene_annotations = []

    occurence_match = occurence_pattern.findall(line)
    occurence: float | None = None
    if occurence_match:
        assert len(occurence_match) == 1
        occurence = float(occurence_match[0])

    return NodeHeader(
        node_id=int(line.split(" ")[0][1:]),
        occurence=occurence,
        metadatas=parsed_gene_annotations,
        sequence=DNA(seq),
        successors=tuple(
            map(lambda e: (int(e[1]), ESign(e[0] + e[2])), find_succ.findall(line)),
        ),
    )


def parse_one_ggcat(spec, seq) -> Tuple[int, str]:
    """Function parsing one line of BCALM file"""
    node_id = int(spec.split(" ")[0])
    return (node_id, seq)


def _buffer_read_data(f, buffsize) -> Iterator[str]:
    remain = ""
    while True:
        read = f.read(buffsize).decode()
        if not read:
            if remain.strip():
                yield remain.strip()
            return
        x = remain + read
        if "\n" not in x:
            remain = x
            continue
        body, remain = x.rsplit("\n", maxsplit=1)
        yield from body.split("\n")


def get_data(filename: Path, buffsize=10**6) -> Iterator[str]:
    if filename.name.endswith(".gz"):
        with gzip.open(filename) as f:
            yield from _buffer_read_data(f, buffsize)
    else:
        with open(filename, "rb") as f:
            yield from _buffer_read_data(f, buffsize)


def stat_bcalm(filename: Path, buffsize=10**6) -> tuple[int, int, int, int]:
    """Read a BCALM format and return the node size, edge size and an estimate for k
    obtained by taking the minimum length of a unitig node
    """
    data = get_data(filename, buffsize=buffsize)
    node_size = 0
    edge_size = 0
    kmer_size = 0
    estimate_k = 63
    for line in data:
        if line and line[0] != ">":
            dna_length = len(line.strip())
            estimate_k = min(estimate_k, dna_length)
            kmer_size += dna_length - estimate_k

            continue
        node_size += 1
        edge_size += line.count("L:")
    return node_size, edge_size, estimate_k, kmer_size


def parse(
    filename: Path,
    buffsize=10**6,
) -> Iterator[NodeHeader]:
    """Function parsing a BCALM file returning an iterator over the parsed value

    Assume the file fits in RAM. More work is needed if it isn't the case.
    """
    data = get_data(filename, buffsize=buffsize)
    values = groupby(enumerate(data), lambda e: e[0] // 2)
    yield from (parse_one(spec, seq) for i, ((_, spec), (_, seq)) in values)


def parse_annotations_for_genes(file_path: Path, k: int) -> Iterator[SubseqData]:
    if file_path.suffix == ".gff":
        with open(file_path) as file:
            for elem in _parse_gff(file, k):
                if elem.type.upper() == "TRANSCRIPT":
                    yield elem
    elif file_path.suffix == ".gtf":
        with open(file_path) as file:
            for elem in _parse_gtf(file, k):
                if elem.type.upper() == "TRANSCRIPT":
                    yield elem
    else:
        raise FormatInputError(
            f"Annotation format extension is: {file_path.suffix}. Should be .gff or .gtf."
        )


def parse_annotations(file_path: Path, k: int) -> Iterator[SubseqData]:
    if file_path.suffix == ".gff":
        with open(file_path) as file:
            yield from _parse_gff(file, k)
    elif file_path.suffix == ".gtf":
        with open(file_path) as file:
            yield from _parse_gtf(file, k)
    else:
        raise FormatInputError(
            f"Annotation format extension is: {file_path.suffix}. Should be .gff or .gtf."
        )


gene_search = re.compile(r'gene_id "([^"]+)"')
transcript_search = re.compile(r'transcript_id "([^"]+)"')


def _parse_gtf(file, k: int) -> Iterator[SubseqData]:
    for line_counter, line in enumerate(file):
        if not line.strip() or line.startswith("#"):
            continue
        fields = line.split("\t")
        match = gene_search.search(line)
        gene_id = match.group(1) if match else None
        match = transcript_search.search(line)
        transcript_id = match.group(1) if match else None
        id_desc = transcript_id or gene_id
        if not id_desc:
            vizitig_logger.error(
                f"ID not found for line {line_counter} of {file}. Make sure your annotation data respect the correct format.",
            )
            continue

        chr, _, feature_type, start, stop, _, str_strand, _, bulk_attributes = fields
        if str_strand not in ["-", "+"]:
            str_strand = None
        attributes = re.sub(r";\s+", ";", re.sub(r";\n", "", bulk_attributes))
        list_attributes = list(filter(bool, map(str.strip, attributes.split(";"))))
        start, stop = int(start), int(stop)
        object_type = feature_type[0].upper() + feature_type[1:]
        if stop - start < k:
            stop = start + k
        yield SubseqData(
            id=id_desc,
            type=object_type,
            list_attr=list_attributes,
            start=start,
            stop=stop,
            gene=gene_id,
            chr=chr,
            strand=str_strand,
        )


id_parser = re.compile(r"ID=([^;]+)")
parent_parser = re.compile(r"Parent=([^;]+)")


def _parse_gff(file, k: int) -> Iterator[SubseqData]:
    for line_counter, line in enumerate(file):
        if not line or line.startswith("#"):
            continue
        id_parsed = id_parser.search(line)
        parent_parsed = parent_parser.search(line)
        id_desc: str | None = None
        if id_parsed is not None:
            id_desc = id_parsed.group(1)
        elif parent_parsed is not None:
            id_desc = parent_parsed.group(1)
        if id_desc is None:
            vizitig_logger.error(
                f"ID not found for line {line_counter} of {file}. Make sure your annotation data respect the correct format.",
            )
            continue

        line = line.strip()
        fields = line.split("\t")

        _, _, feature_type, start, stop, _, strand, _, _ = fields
        _, _, feature_type, start, stop, _, _, _, bulk_attributes = fields
        attributes = re.sub(r";\s+", ";", re.sub(r";\n", "", bulk_attributes))
        list_attributes = attributes.split(";")
        start, stop = int(start), int(stop)
        object_type = feature_type[0].upper() + feature_type[1:]
        if stop - start < k:
            stop = start + k

        yield SubseqData(
            id=id_desc,
            type=object_type,
            list_attr=list_attributes,
            start=start,
            stop=stop,
        )


def parse_reference_sequence_for_full_annotation(
    refseq: Path,
    metadatas: Iterable[SubseqData],
    k: int,
    logger_name: str = "vizitig",
) -> Iterator[Tuple[SubseqData, DNA]]:
    logger = logging.getLogger(f"{logger_name}.refparsing")

    # This implementation will load the reference sequence in RAM.

    with open(refseq) as file:
        sequence: str = file.read().replace("\n", "")

        for i, meta in enumerate(metadatas):
            try:
                dna: str = DNA.from_str(sequence[meta.start : meta.stop])
                for d in dna:
                    yield (meta, d)
            except IndexError:
                logger.warning("""
                            Metadata {} of type {} has starts and stop that are not included in the provided reference sequence.
                            If you are not working with a subset, check your reference file corresponds to your annotation file.
                            """)
        logger.info("Finished")



import argparse
import os
import shutil
import tempfile
from contextlib import contextmanager
from pathlib import Path
from typing import Any

import networkdisk as nd
from networkdisk.sqlite import DiGraph, Graph
from networkdisk.sqlite import sqlitedialect as sqlited
from tabulate import tabulate, tabulate_formats

from vizitig.cli import subparsers
from vizitig.errors import FileExists, NameAlreadyTaken, NoGraphException
from vizitig.metadata import GraphMetadata
from vizitig.paths import (
    graph_path_name,
    index_path_name,
    graphs_log,
    graphs_path,
    log_path_name,
)
from vizitig.types import Zero
from vizitig.utils import (
    SubLog,
    new_gid,
    number_fmt,
    sizeof_fmt,
)
from vizitig.utils import (
    vizitig_logger as logger,
)

from vizitig.env_var import VIZITIG_TMP_DIR

graphs = {}


def graphs_list() -> list[str]:
    # TODO: this is probably not portable
    graphs = set([f.stem for f in graphs_path.glob("*.db") if f.stem[0] != "."])
    return sorted(graphs)


def clean_log():
    if graphs_log.exists():
        shutil.rmtree(graphs_log)


def are_kmer_ingested(gname) -> bool:
    """We say yes if at least one value greater than 0 exists"""
    G = get_graph(gname)
    if not hasattr(G, "kmer_ingested"):
        try:
            next(iter(G.find_all_nodes(lambda e: e.gt(Zero()))))
            G.kmer_ingested = True
        except StopIteration:
            G.kmer_ingested = False
    return G.kmer_ingested


def create_graph(
    graph_name: str,
    k: int,
    output_path: Path,
    size: int,
    kmer_size: int,
    edge_size: int,
    sql_logger: bool = False,
    directed_edges: bool = True,
) -> DiGraph | Graph:
    GM = GraphMetadata(
        name=graph_name,
        k=k,
        size=size,
        kmer_size=kmer_size,
        edge_size=edge_size,
        gid=new_gid(),
    )
    graph_factory = Graph
    if directed_edges:
        graph_factory = DiGraph
    G = graph_factory(
        db=output_path,
        schema=dict(
            node="INT",
            node_datakey=f"VIZI_{GM.gid}",
            node_datavalue="TEXT",
        ),
        sql_logger=sql_logger,
        name="kmer_graph",
    )
    MetadataGraph = DiGraph(db=G.helper, sql_logger=sql_logger, name="metadata_graph")
    MetadataGraph.graph["size"] = 0
    G.metadata = GM
    G.graph_metadata = MetadataGraph
    GM.commit_to_graph(G=G, set_offset=False)
    return G


def regenerate_schema(graph):
    some_G = type(graph)(
        schema=dict(
            node="INT",
            node_datakey=f"VIZI_{graph.metadata.gid}",
            node_datavalue="TEXT",
        ),
    )
    graph.master.delete_graph(masterid=1)
    graph.master.save_graph(some_G, masterid=1)


def rename_graph(old_name: str, new_name: str, replace: bool = False) -> None:
    L = graphs_list()
    if old_name not in L:
        raise NoGraphException(old_name)
    if new_name in L and not replace:
        raise NameAlreadyTaken(new_name)
    shutil.move(graph_path_name(old_name), graph_path_name(new_name))
    shutil.move(
        index_path_name(old_name, create=False), index_path_name(new_name, create=False)
    )
    G = get_graph(new_name)
    G.graph["name"] = new_name


def delete_graph(name: str) -> None:
    L = graphs_list()
    if name not in L:
        raise NoGraphException(name)
    shutil.rmtree(index_path_name(name))
    os.unlink(graph_path_name(name))


def add_vizitig_graph(
    path: Path,
    name: str | None,
    replace: bool = False,
    check_compatibility: bool = False,
    copy=False,
) -> None:
    """Add an already existing graph to vizitig.
    Some issue with gid:
        to avoid issues, each graph should have a distinct gid.
        This is used to encode/decode metadata as networkdisk is too limited for now
        to do that in a flexible way.

        Hence, we need to regenerate the schema with type which depends of this gid.
    """
    if not path.exists():
        raise ValueError(f"Invalid {path}")
    if name is None:
        name = path.stem
    if name == "":
        raise ValueError("Graph name should be non empty")
    if name in graphs_list() and not replace:
        raise FileExists(f"{name} already exists")
    G = get_graph_from_path(path)
    GraphMetadata.set_metadata(
        G,
        check_compatibility=check_compatibility,
    )  # check compatibility mostly

    if copy:
        with tempfile.NamedTemporaryFile(prefix=VIZITIG_TMP_DIR, delete=False) as f:
            logger.info(f"Copy {path} to {f.name}")
            shutil.copy(path, f.name)  # this should happen info to log file eventually
            logger.info(f"mv {f.name} to {graph_path_name(name)}")
            shutil.move(f.name, graph_path_name(name))
    else:
        shutil.move(path, graph_path_name(name))

    new_G = get_graph_from_path(graph_path_name(name))
    GraphMetadata.set_metadata(new_G)
    new_G.metadata.gid = new_gid()  # refresh gid to avoid collision when dl/copying etc
    new_G.metadata.model_post_init(None)
    regenerate_schema(new_G)
    new_G.metadata.commit_to_graph()


def get_graph_from_path(
    path: Path,
    sql_logger=False,
    check_compatibility=True,
) -> DiGraph | Graph:
    master = sqlited.master.MasterGraphs(db=path)
    ## UGLY HACK
    # we can't know the gid of the graph before loading it
    # we can't load the graph without the gid because stored in the graph schema
    # we catch the gid with the key_error raise in the encoder function, set dummy function
    # and then build the metdata.
    key = None
    try:
        G = master.load_graph(name="kmer_graph")
    except KeyError as K:
        key = K.args[0]
        nd.utils.serialize.encoderFunctions[key] = (lambda e: e, lambda e: e)
        G = master.load_graph(name="kmer_graph")
    G.metadata_graph = master.load_graph(name="metadata_graph")
    return G


def get_graph(name: str, sql_logger=False, check_compatibility=True) -> DiGraph | Graph:
    if name not in graphs:
        L = graphs_list()
        if name not in L:
            raise NoGraphException(name)
        if not name:
            raise ValueError
        G = graphs[name] = get_graph_from_path(graph_path_name(name))
        GraphMetadata.set_metadata(
            G, check_compatibility=check_compatibility, name=name
        )  # this should be safe to remove
        # once all graph in the wild have a metadata.name set
        # this is a quickfix to avoid a break in compatibility of old graph
    else:
        G = graphs[name]
    return G


def reload_graph(name: str) -> None:
    G = get_graph(name)
    G.helper.close()
    del graphs[name]
    del G
    return get_graph(name)


def graph_info(name: str, human_size: bool = False) -> dict[str, Any]:
    from vizitig.index import index_info  # avoiding circular import

    G = get_graph(name)
    f = graph_path_name(name)
    d = G.metadata.model_dump()
    d.update(
        dict(
            name=f.stem,
            file_size=f.stat().st_size,
            path=str(f),
            log_exists=log_path_name(name).exists(),
            index=index_info(name),
        ),
    )
    if human_size:
        d["file_size"] = sizeof_fmt(d["file_size"])
        d["node nb"] = number_fmt(d.pop("size"))
        d["edge nb"] = number_fmt(d.pop("edge_size"))
        d["index"] = [idx.model_dump() for idx in d["index"]]
        for idx in d["index"]:
            idx["size"] = sizeof_fmt(idx["size"])

    return d


def pretty_info(d: dict[str, Any]) -> dict[str, Any]:
    pretty_d = dict()
    for k, v in d.items():
        if isinstance(v, list):
            if len(v) > 3:
                pretty_d[k] = ", ".join(map(repr, v[:3] + [f"... ({len(v)})"]))
            else:
                pretty_d[k] = ", ".join(map(repr, v))
        else:
            pretty_d[k] = v
    for key in ("types_list", "filter_list", "log_exists"):
        pretty_d.pop(key)
    pretty_d["path"] = str(pretty_d["path"])
    pretty_d["index"] = ", ".join(
        map(lambda e: f"{e['type']}({e['size']}, k={e['k']})", d["index"]),
    )
    pretty_d["Viz Vers."] = pretty_d.pop("vizitig_version")
    pretty_d.pop("gid")
    return pretty_d


def main(args: argparse.Namespace) -> None:
    if args.format == "json":
        from json import dumps as jsondumps

        data = [graph_info(f) for f in graphs_list()]
        for d in data:
            d["index"] = [idx.model_dump() for idx in d["index"]]
        print(jsondumps(data))
    elif args.format in tabulate_formats:
        data = [pretty_info(graph_info(f, human_size=True)) for f in graphs_list()]
        print(tabulate(data, headers="keys", tablefmt=args.format))
    else:
        raise NotImplementedError(args.format)


@contextmanager
def GraphLogger(name: str, message: str | None = None):
    path = log_path_name(name)
    if message is None:
        message = name
    try:
        with SubLog(message, file=path):
            yield path
    finally:
        if path.exists():
            path.unlink()


OutputFormat = ["human", "json"]

parser = subparsers.add_parser(
    "info",
    help="get information about ingested graphs",
)

parser.add_argument(
    "--format",
    help=f"Output format, default is human readable text, choice: {OutputFormat}",
    metavar="output",
    type=str,
    choices=OutputFormat,
    default="simple",
)
parser.set_defaults(func=main)


def main_rename(arg: argparse.Namespace):
    rename_graph(arg.old_name, arg.new_name, replace=arg.replace)


parser = subparsers.add_parser(
    "rename",
    help="Rename a graph",
)

parser.add_argument(
    "old_name",
    help="Current graph name",
    type=str,
)

parser.add_argument(
    "new_name",
    help="New graph name",
    metavar="old_graph",
    type=str,
)

parser.add_argument(
    "-r",
    "--replace",
    help="Replace existing graph",
    action="store_true",
    default=False,
)

parser.set_defaults(func=main_rename)


def main_import(arg: argparse.Namespace):
    add_vizitig_graph(arg.path, name=arg.name, replace=arg.replace)


parser = subparsers.add_parser(
    "add",
    help="Add an already built Vizitig Graph",
)

parser.add_argument(
    "path",
    help="Path to the file",
    type=Path,
)

parser.add_argument(
    "-n",
    "--name",
    help="A name for the graph (default is the file name)",
    type=str,
)

parser.add_argument(
    "-r",
    "--replace",
    help="Replace existing graph",
    action="store_true",
    default=False,
)

parser.set_defaults(func=main_import)


parser = subparsers.add_parser(
    "rm",
    help="Remove an already built Vizitig Graph",
)

parser.add_argument(
    "name",
    help=f"The name of the graph: {graphs_list()}",
    type=str,
    metavar="name",
    choices=graphs_list(),
)


def main_delete(arg: argparse.Namespace):
    delete_graph(arg.name)


parser.set_defaults(func=main_delete)



import sqlite3
import tempfile
import shutil
from pathlib import Path
from typing import Iterable, Iterator, Tuple
from typing_extensions import Self
from vizitig.env_var import VIZITIG_TMP_DIR

from vizitig.index.classes import Shard, SetShard
from vizitig.types import Kmer, encode_kmer, decode_kmer
from vizitig.utils import SubLog
from vizitig.utils import vizitig_logger as logger


class BaseSQLite:
    priority = 3
    k: int

    @property
    def connector(self):
        return sqlite3.connect(self._path)

    def __len__(self) -> int:
        with self.connector as db:
            c = db.cursor()
            return next(iter(c.execute("SELECT count FROM kmer_size")))[0]

    def __iter__(self) -> Iterator[Tuple[Kmer]]:
        with self.connector as db:
            c = db.cursor()
            yield from (
                decode_kmer(kmer, self.k)
                for (kmer,) in c.execute("SELECT kmer FROM kmer_idx")
            )


class SQLiteSet(BaseSQLite, SetShard):
    @classmethod
    def build_kmer(
        cls,
        path: Path,
        shard_idx: int,
        shard_number: int,
        kmer_iter: Iterator[Kmer],
        k: int,
    ) -> Self:
        if k < 32:
            kmer_type = "INTEGER"
        else:
            kmer_type = "BLOB"
        if path.exists():
            logger.warning("Index already exists, erasing and rebuilding")
            path.unlink()

        tmp_fd, _tmp_path = tempfile.mkstemp(prefix=VIZITIG_TMP_DIR)
        tmp_path = Path(_tmp_path)
        with sqlite3.connect(tmp_path) as db:
            db.execute(
                f"CREATE TABLE kmer_idx(kmer {kmer_type} PRIMARY KEY) WITHOUT ROWID",
            )
            c = db.cursor()
            c.executemany(
                "INSERT OR IGNORE INTO kmer_idx(kmer) VALUES (?) ",
                (
                    (encode_kmer(kmer, k),)
                    for kmer in kmer_iter
                    if (hash(kmer) % shard_number == shard_idx)
                ),
            )

            logger.info("Counting kmers")
            c.execute(
                "CREATE TABLE kmer_size AS SELECT count(1) as count FROM kmer_idx",
            )
            c.close()
        logger.info(f"moving from {tmp_path} to {path}")
        shutil.move(tmp_path, path)
        return cls(path, shard_idx, shard_number, k)

    def __contains__(self, kmer: Kmer) -> bool:
        if hash(kmer) % self._shard_number != self._shard_idx:
            return False

        with self.connector as db:
            c = db.cursor()
            try:
                next(
                    iter(
                        c.execute(
                            "SELECT 1 FROM kmer_idx WHERE kmer=?",
                            (encode_kmer(kmer, self.k),),
                        ),
                    ),
                )[0]
                return True
            except StopIteration:
                return False


class SQLiteIndex(BaseSQLite, Shard):
    Set = SQLiteSet

    @classmethod
    def build_kmer(
        cls,
        path: Path,
        shard_idx: int,
        shard_number: int,
        kmer_iter: Iterator[Tuple[Kmer, int]],
        k: int,
    ) -> Self:
        if k < 32:
            kmer_type = "INTEGER"
        else:
            kmer_type = "BLOB"
        if path.exists():
            logger.warning("Index already exists, erasing and rebuilding")
            path.unlink()

        tmp_fd, _tmp_path = tempfile.mkstemp(prefix=VIZITIG_TMP_DIR)
        tmp_path = Path(_tmp_path)
        with sqlite3.connect(tmp_path) as db:
            db.execute(
                f"CREATE TABLE kmer_idx(kmer {kmer_type}, val INTEGER, PRIMARY KEY (kmer, val)) WITHOUT ROWID",
            )
            c = db.cursor()
            c.executemany(
                "INSERT OR IGNORE INTO kmer_idx(kmer, val) VALUES (?, ?) ",
                (
                    (encode_kmer(kmer, k), val)
                    for kmer, val in kmer_iter
                    if (hash(kmer) % shard_number == shard_idx)
                ),
            )

            logger.info("Counting kmers")
            c.execute(
                "CREATE TABLE kmer_size AS SELECT count(1) as count FROM kmer_idx",
            )
            c.close()
        logger.info(f"moving from {tmp_path} to {path}")
        shutil.move(tmp_path, path)
        return cls(path, shard_idx, shard_number, k)

    def __getitem__(self, kmer: Kmer) -> int:
        if hash(kmer) % self._shard_number != self._shard_idx:
            raise KeyError(kmer)

        with self.connector as db:
            c = db.cursor()
            try:
                return next(
                    iter(
                        c.execute(
                            "SELECT val FROM kmer_idx WHERE kmer=?",
                            (encode_kmer(kmer, self.k),),
                        ),
                    ),
                )[0]
            except StopIteration:
                raise KeyError(kmer)
            finally:
                c.close()

    def get_all(self, kmer: Kmer) -> Iterable[int]:
        if hash(kmer) % self._shard_number != self._shard_idx:
            raise KeyError(kmer)

        with self.connector as db:
            c = db.cursor()
            L = list(
                c.execute(
                    "SELECT val FROM kmer_idx WHERE kmer=?",
                    (encode_kmer(kmer, self.k),),
                ),
            )
            c.close()
            return L

    def join(
        self,
        other: Iterable[Tuple[Kmer, int]],
    ) -> Iterable[Tuple[int, int]]:
        """First we compute an offset for each vizikey.
        Then we insert in a temp table kmer, vizikey_offset
        Then we perform in DB the join.

        This function should be safe to be used with multiprocessing.
        """

        def build_iterator() -> Iterable[Tuple[Kmer, int]]:
            for kmer, value in other:
                yield (encode_kmer(kmer, self.k), value)

        with SubLog("SQLite Join"):
            with self.connector as db:
                c = db.cursor()
                c.execute(
                    "CREATE TEMPORARY TABLE metadata_tmp(kmer INTEGER, key INTEGER, PRIMARY KEY (kmer, key)) WITHOUT ROWID",
                )
                c.executemany(
                    "INSERT OR IGNORE INTO metadata_tmp VALUES (?, ?)", build_iterator()
                )
                logger.info("starting the join")
                it = c.execute("""
SELECT DISTINCT val, key
FROM 
    kmer_idx INNER JOIN metadata_tmp ON kmer_idx.kmer = metadata_tmp.kmer 
""")
                yield from it
                c.close()

    def join_index(self, other: Self) -> Iterable[Tuple[int, int]]:
        assert self.k == other.k

        with self.connector as db:
            c = db.cursor()
            c.execute("ATTACH ? as other", (str(other.path),))
            it = c.execute("""
SELECT DISTINCT A.val, B.val from kmer_idx as A INNER JOIN other.kmer_idx as B
    ON A.kmer = B.kmer
""")
            yield from it
            c.close()

    def intersection(self, other: Iterable[Kmer]) -> Iterable[int]:
        with SubLog("SQLite intersection"):
            with self.connector as db:
                c = db.cursor()
                c.execute(
                    "CREATE TEMPORARY TABLE kmer_tmp(kmer INTEGER PRIMARY KEY) WITHOUT ROWID"
                )
                c.executemany(
                    "INSERT OR IGNORE INTO kmer_tmp VALUES (?)",
                    ((encode_kmer(e, self.k),) for e in other),
                )

                logger.info("starting the intersection")
                yield from (
                    t[0]
                    for t in c.execute("""
SELECT DISTINCT val
FROM 
    kmer_idx INNER JOIN kmer_tmp ON kmer_idx.kmer = kmer_tmp.kmer 
""")
                )
                c.close()

    def intersection_index(self, other: SQLiteSet) -> Iterable[int]:
        assert self.k == other.k

        with self.connector as db:
            c = db.cursor()
            c.execute("ATTACH ? as other", (str(other.path),))
            it = c.execute("""
SELECT DISTINCT A.val from kmer_idx as A INNER JOIN other.kmer_idx as B
    ON A.kmer = B.kmer
""")
            yield from (e[0] for e in it)
            c.close()



import shutil
import tempfile

from vizitig.env_var import VIZITIG_TMP_DIR
from vizitig.index.classes import Shard, SetShard
from vizitig.utils import SubLog
from vizitig.types import Kmer, DNA
from vizitig.utils import vizitig_logger as logger
from pathlib import Path
from typing import Iterator, Tuple, Iterable, Type
from typing_extensions import Self


try:
    from vizibridge.kmer_index import KmerIndex, KmerSet

    class BaseRust:
        priority = 4
        k: int

        @property
        def base_idx(self):
            if not hasattr(self, "_base_idx"):
                self._base_idx = self.base_cls(Path(self.path), self.k)
            return self._base_idx

        def __len__(self) -> int:
            return len(self.base_idx)

        def __iter__(self) -> Iterator[Kmer]:
            last = None
            for kmer in self.base_idx:
                if kmer == last:
                    continue
                last = kmer
                yield Kmer(kmer, self.k)

    class RustSet(BaseRust, SetShard):
        @property
        def base_cls(self) -> Type:
            return KmerSet

        @classmethod
        def build_kmer(
            cls,
            path: Path,
            shard_idx: int,
            shard_number: int,
            kmer_iter: Iterator[Kmer],
            k: int,
        ) -> Self:
            with SubLog(f"Rust Build KmerSet({path})"):
                tmp_fd, _tmp_path = tempfile.mkstemp(prefix=VIZITIG_TMP_DIR)
                open(tmp_fd).close()
                tmp_path = Path(_tmp_path)
                tmp_path.unlink()  # UGLY: if the file exists, rust code will complains
                index_iter = (
                    kmer
                    for kmer in kmer_iter
                    if (hash(kmer) % shard_number == shard_idx)
                )
                logger.info(f"Building index in {tmp_path}")
                KmerSet.build(index_iter, tmp_path, k)
                logger.info(f"file size: {tmp_path.stat().st_size}")
                logger.info(f"moving from {tmp_path} to {path}")
                shutil.move(tmp_path, path)
            return cls(path, shard_idx, shard_number, k)

        @classmethod
        def build_dna(
            cls,
            path: Path,
            shard_idx: int,
            shard_number: int,
            dna_iter: Iterator[DNA],
            k: int,
        ) -> Self:
            with SubLog(f"Rust Build KmerSet(DNA)({path})"):
                tmp_fd, _tmp_path = tempfile.mkstemp(prefix=VIZITIG_TMP_DIR)
                open(tmp_fd).close()
                tmp_path = Path(_tmp_path)
                tmp_path.unlink()  # UGLY: if the file exists, rust code will complains
                logger.info(f"Building index in {tmp_path}")
                KmerSet.build_dna(dna_iter, tmp_path, k, shard_idx, shard_number)
                logger.info(f"stat: {tmp_path.stat().st_size}")
                logger.info(f"moving from {tmp_path} to {path}")
                shutil.move(tmp_path, path)
            return cls(path, shard_idx, shard_number, k)

        def __contains__(self, kmer: Kmer) -> bool:
            return kmer in self.base_idx

    class RustIndex(BaseRust, Shard):
        priority = 4
        Set = RustSet

        @property
        def base_cls(self) -> Type:
            return KmerIndex

        @classmethod
        def build_kmer(
            cls,
            path: Path,
            shard_idx: int,
            shard_number: int,
            kmer_iter: Iterator[Tuple[Kmer, int]],
            k: int,
        ) -> Self:
            with SubLog(f"Rust Build Kmer({path})"):
                tmp_fd, _tmp_path = tempfile.mkstemp(prefix=VIZITIG_TMP_DIR)
                open(tmp_fd).close()
                tmp_path = Path(_tmp_path)
                tmp_path.unlink()  # UGLY: if the file exists, rust code will complains
                index_iter = (
                    (kmer, val)
                    for kmer, val in kmer_iter
                    if (hash(kmer) % shard_number == shard_idx)
                )
                logger.info(f"Building index in {tmp_path}")
                KmerIndex.build(index_iter, tmp_path, k)
                logger.info(f"stat: {tmp_path.stat()}")
                logger.info(f"moving from {tmp_path} to {path}")
                shutil.move(tmp_path, path)
            return cls(path, shard_idx, shard_number, k)

        @classmethod
        def build_dna(
            cls,
            path: Path,
            shard_idx: int,
            shard_number: int,
            dna_iter: Iterator[Tuple[DNA, int]],
            k: int,
        ) -> Self:
            with SubLog(f"Rust Build DNA ({path})"):
                tmp_fd, _tmp_path = tempfile.mkstemp(prefix=VIZITIG_TMP_DIR)
                open(tmp_fd).close()
                tmp_path = Path(_tmp_path)
                tmp_path.unlink()  # UGLY: if the file exists, rust code will complains
                logger.info(f"Building index in {tmp_path}")
                KmerIndex.build_dna(dna_iter, tmp_path, k, shard_idx, shard_number)
                logger.info(f"stat: {tmp_path.stat().st_size}")
                logger.info(f"moving from {tmp_path} to {path}")
                shutil.move(tmp_path, path)
            return cls(path, shard_idx, shard_number, k)

        def __getitem__(self, kmer: Kmer) -> int:
            return self.base_idx[kmer]

        def get_all(self, kmer: Kmer) -> Iterable[int]:
            return self.base_idx.get_all(kmer.base_type)

        def join(
            self, kmer_iter: Iterator[Tuple[Kmer, int]]
        ) -> Iterable[Tuple[int, int]]:
            path = self.get_temp_file()
            with SubLog(f"Rust Join Kmer-Iter({path})"):
                try:
                    kmer_index_entry_it = (
                        (kmer.base_type, val) for kmer, val in kmer_iter
                    )
                    yield from self.base_idx.join(kmer_index_entry_it, path)
                finally:
                    path.unlink()

        def join_index(self, other: "RustIndex") -> Iterable[Tuple[int, int]]:
            path = self.get_temp_file()
            with SubLog(f"Rust Join Index({path})"):
                try:
                    return self.base_idx.join_index(other.base_idx, path)
                finally:
                    path.unlink()

        def intersection(self, kmer_iter: Iterator[Kmer]) -> Iterable[int]:
            path = self.get_temp_file()
            with SubLog(f"Rust Intersection Kmer-Iter({path})"):
                try:
                    kmer_base = (kmer.base_type for kmer in kmer_iter)
                    return self.base_idx.intersection(kmer_base, path)
                finally:
                    path.unlink()

        def intersection_index(self, other: RustSet) -> Iterable[int]:
            path = self.get_temp_file()
            with SubLog(f"Rust Intersecton Index({path})"):
                try:
                    return self.base_idx.intersection_index(other.base_idx, path)
                finally:
                    path.unlink()


except ImportError as E:  # This could happen with older version of vizibridge
    logger.warning(
        f"Rust Index is not installed, fallback to less efficient index ({E})"
    )
    ...
